@@ -23,7 +23,6 @@
  *   Boston, MA 02110-1301 USA
  */
 
-#define _GNU_SOURCE
 #include <stdio.h>
 #include <netdb.h>
 #include <unistd.h>
@@ -51,7 +50,7 @@ static int append_opt(char *body, int bodylen, char *opt, char *name)
 	}
 
 	while (*opt) {
-		if (isalnum(*opt)) {
+		if (isalnum((int)(unsigned char)*opt)) {
 			if (len >= bodylen - 1)
 				return -ENOSPC;
 			body[len++] = *opt;
@@ -69,7 +68,7 @@ static int append_opt(char *body, int bodylen, char *opt, char *name)
 	body[len++] = '=';
 
 	while (name && *name) {
-		if (isalnum(*name)) {
+		if (isalnum((int)(unsigned char)*name)) {
 			if (len >= bodylen - 1)
 				return -ENOSPC;
 			body[len++] = *name;
@@ -121,7 +120,7 @@ static int parse_auth_choice(struct openconnect_info *vpninfo, struct oc_auth_fo
 	opt->form.label = (char *)xmlGetProp(xml_node, (unsigned char *)"label");
 
 	if (!opt->form.name) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Form choice has no name\n");
+		vpn_progress(vpninfo, PRG_ERR, _("Form choice has no name\n"));
 		free(opt);
 		return -EINVAL;
 	}
@@ -184,13 +183,15 @@ static int parse_form(struct openconnect_info *vpninfo, struct oc_auth_form *for
 			continue;
 		}
 		if (strcmp((char *)xml_node->name, "input")) {
-			vpninfo->progress(vpninfo, PRG_TRACE, "name %s not input\n", xml_node->name);
+			vpn_progress(vpninfo, PRG_TRACE,
+				     _("name %s not input\n"), xml_node->name);
 			continue;
 		}
 
 		input_type = (char *)xmlGetProp(xml_node, (unsigned char *)"type");
 		if (!input_type) {
-			vpninfo->progress(vpninfo, PRG_INFO, "No input type in form\n");
+			vpn_progress(vpninfo, PRG_INFO,
+				     _("No input type in form\n"));
 			continue;
 		}
 
@@ -201,7 +202,8 @@ static int parse_form(struct openconnect_info *vpninfo, struct oc_auth_form *for
 
 		input_name = (char *)xmlGetProp(xml_node, (unsigned char *)"name");
 		if (!input_name) {
-			vpninfo->progress(vpninfo, PRG_INFO, "No input name in form\n");
+			vpn_progress(vpninfo, PRG_INFO,
+				     _("No input name in form\n"));
 			free(input_type);
 			continue;
 		}
@@ -223,9 +225,9 @@ static int parse_form(struct openconnect_info *vpninfo, struct oc_auth_form *for
 		else if (!strcmp(input_type, "password"))
 			opt->type = OC_FORM_OPT_PASSWORD;
 		else {
-			vpninfo->progress(vpninfo, PRG_INFO,
-					  "Unknown input type %s in form\n",
-					  input_type);
+			vpn_progress(vpninfo, PRG_INFO,
+				     _("Unknown input type %s in form\n"),
+				     input_type);
 			free(input_type);
 			free(input_name);
 			free(input_label);
@@ -244,7 +246,7 @@ static int parse_form(struct openconnect_info *vpninfo, struct oc_auth_form *for
 		*p = opt;
 	}
 
-	vpninfo->progress(vpninfo, PRG_TRACE, "Fixed options give %s\n", body);
+	vpn_progress(vpninfo, PRG_TRACE, _("Fixed options give %s\n"), body);
 
 	return 0;
 }
@@ -255,7 +257,8 @@ static int process_auth_form(struct openconnect_info *vpninfo,
 static char *xmlnode_msg(xmlNode *xml_node)
 {
 	char *fmt = (char *)xmlNodeGetContent(xml_node);
-	char *result, *param1, *param2, *pct;
+	char *result, *params[2], *pct;
+	int len;
 	int nr_params = 0;
 
 	if (!fmt || !fmt[0]) {
@@ -263,34 +266,48 @@ static char *xmlnode_msg(xmlNode *xml_node)
 		return NULL;
 	}
 
-	for (pct = strchr(fmt, '%'); pct;
+	len = strlen(fmt) + 1;
+	
+	params[0] = (char *)xmlGetProp(xml_node, (unsigned char *)"param1");
+	if (params[0])
+		len += strlen(params[0]);
+	params[1] = (char *)xmlGetProp(xml_node, (unsigned char *)"param2");
+	if (params[1])
+		len += strlen(params[1]);
+
+	result = malloc(len);
+	if (!result) {
+		result = fmt;
+		goto out;
+	}
+
+	strcpy(result, fmt);
+	free (fmt);
+
+	for (pct = strchr(result, '%'); pct;
 	     (pct = strchr(pct, '%'))) {
+		int paramlen;
+
 		/* We only cope with '%s' */
 		if (pct[1] != 's')
-			return fmt;
-		pct++;
-		nr_params++;
+			goto out;
+
+		if (params[nr_params]) {
+			paramlen = strlen(params[nr_params]);
+			/* Move rest of fmt string up... */
+			memmove(pct - 1 + paramlen, pct + 2, strlen(pct) - 1);
+			/* ... and put the string parameter in where the '%s' was */
+			memcpy(pct, params[nr_params], paramlen);
+			pct += paramlen;
+		} else
+			pct++;
+
+		if (++nr_params == 2)
+			break;
 	}
-	if (!nr_params)
-		return fmt;
-
-	param1 = (char *)xmlGetProp(xml_node, (unsigned char *)"param1");
-	param2 = (char *)xmlGetProp(xml_node, (unsigned char *)"param2");
-
-	if (nr_params && !param1)
-		return result;
-	if (nr_params > 1 && !param2)
-		return result;
-	if (nr_params > 2)
-		return result;
-
-	if (asprintf(&result, fmt, param1, param2) < 0)
-		result = fmt;
-	else 
-		free(fmt);
-
-	free(param1);
-	free(param2);
+ out:
+	free(params[0]);
+	free(params[1]);
 	return result;
 }
 
@@ -301,8 +318,8 @@ static char *xmlnode_msg(xmlNode *xml_node)
  *  = 2, when form indicates that login was already successful
  */
 int parse_xml_response(struct openconnect_info *vpninfo, char *response,
-		       char *request_body, int req_len, char **method,
-		       char **request_body_type)
+		       char *request_body, int req_len, const char **method,
+		       const char **request_body_type)
 {
 	struct oc_auth_form *form;
 	xmlDocPtr xml_doc;
@@ -316,15 +333,18 @@ int parse_xml_response(struct openconnect_info *vpninfo, char *response,
 
 	xml_doc = xmlReadMemory(response, strlen(response), "noname.xml", NULL, 0);
 	if (!xml_doc) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Failed to parse server response\n");
-		vpninfo->progress(vpninfo, PRG_TRACE, "Response was:%s\n", response);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to parse server response\n"));
+		vpn_progress(vpninfo, PRG_TRACE,
+			     _("Response was:%s\n"), response);
 		free(form);
 		return -EINVAL;
 	}
 
 	xml_node = xmlDocGetRootElement(xml_doc);
 	if (xml_node->type != XML_ELEMENT_NODE || strcmp((char *)xml_node->name, "auth")) {
-		vpninfo->progress(vpninfo, PRG_ERR, "XML response has no \"auth\" root node\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("XML response has no \"auth\" root node\n"));
 		ret = -EINVAL;
 		goto out;
 	}
@@ -336,7 +356,8 @@ int parse_xml_response(struct openconnect_info *vpninfo, char *response,
 	}
 
 	if (vpninfo->nopasswd) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Asked for password but '--no-passwd' set\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Asked for password but '--no-passwd' set\n"));
 		ret = -EPERM;
 		goto out;
 	}
@@ -360,9 +381,9 @@ int parse_xml_response(struct openconnect_info *vpninfo, char *response,
 			form->action = (char *)xmlGetProp(xml_node, (unsigned char *)"action");
 			if (!form->method || !form->action || 
 			    strcasecmp(form->method, "POST") || !form->action[0]) {
-				vpninfo->progress(vpninfo, PRG_ERR,
-						  "Cannot handle form method='%s', action='%s'\n",
-						  form->method, form->action);
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Cannot handle form method='%s', action='%s'\n"),
+					     form->method, form->action);
 				ret = -EINVAL;
 				goto out;
 			}
@@ -407,15 +428,15 @@ int parse_xml_response(struct openconnect_info *vpninfo, char *response,
 	}
 	if (!form->opts) {
 		if (form->message)
-			vpninfo->progress(vpninfo, PRG_INFO, "%s\n", form->message);
+			vpn_progress(vpninfo, PRG_INFO, "%s\n", form->message);
 		if (form->error)
-			vpninfo->progress(vpninfo, PRG_ERR, "%s\n", form->error);
+			vpn_progress(vpninfo, PRG_ERR, "%s\n", form->error);
 		ret = -EPERM;
 		goto out;
 	}
 
 	if (vpninfo->process_auth_form)
-		ret = vpninfo->process_auth_form(vpninfo, form);
+		ret = vpninfo->process_auth_form(vpninfo->cbdata, form);
 	else
 		ret = process_auth_form(vpninfo, form);
 	if (ret)
@@ -481,7 +502,7 @@ static int process_auth_form(struct openconnect_info *vpninfo,
 	choice_resp[0] = 0;
 
 	if (!ui) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Failed to create UI\n");
+		vpn_progress(vpninfo, PRG_ERR, _("Failed to create UI\n"));
 		return -EINVAL;
 	}
 	if (form->banner) {
@@ -523,9 +544,9 @@ static int process_auth_form(struct openconnect_info *vpninfo,
 					}
 				}
 				if (!opt->value)
-					vpninfo->progress(vpninfo, PRG_ERR,
-							  "Auth choice \"%s\" not available\n",
-							  vpninfo->authgroup);
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Auth choice \"%s\" not available\n"),
+						     vpninfo->authgroup);
 			}
 			if (!opt->value && select_opt->nr_choices == 1) {
 				choice = &select_opt->choices[0];
@@ -604,7 +625,7 @@ static int process_auth_form(struct openconnect_info *vpninfo,
 		goto out_ui;
 	case -1:
 		/* error */
-		vpninfo->progress(vpninfo, PRG_ERR, "Invalid inputs\n");
+		vpn_progress(vpninfo, PRG_ERR, _("Invalid inputs\n"));
 		ret = -EINVAL;
 	out_ui:
 		UI_free(ui);
@@ -626,9 +647,9 @@ static int process_auth_form(struct openconnect_info *vpninfo,
 			}
 		}
 		if (!select_opt->form.value) {
-			vpninfo->progress(vpninfo, PRG_ERR,
-					  "Auth choice \"%s\" not valid\n",
-					  choice_resp);
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Auth choice \"%s\" not valid\n"),
+				     choice_resp);
 			return -EINVAL;
 		}
 	}
