@@ -23,7 +23,6 @@
  *   Boston, MA 02110-1301 USA
  */
 
-#define _GNU_SOURCE
 #include <netdb.h>
 #include <unistd.h>
 #include <fcntl.h>
@@ -59,7 +58,8 @@ static int http_add_cookie(struct openconnect_info *vpninfo,
 	if (*value) {
 		new = malloc(sizeof(*new));
 		if (!new) {
-			vpninfo->progress(vpninfo, PRG_ERR, "No memory for allocating cookies\n");
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("No memory for allocating cookies\n"));
 			return -ENOMEM;
 		}
 		new->next = NULL;
@@ -113,7 +113,8 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 
  cont:
 	if (openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)) < 0) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Error fetching HTTPS response\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Error fetching HTTPS response\n"));
 		return -EINVAL;
 	}
 
@@ -121,24 +122,27 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		closeconn = 1;
 	
 	if ((!closeconn && strncmp(buf, "HTTP/1.1 ", 9)) || !(*result = atoi(buf+9))) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Failed to parse HTTP response '%s'\n", buf);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to parse HTTP response '%s'\n"), buf);
 		return -EINVAL;
 	}
 
-	vpninfo->progress(vpninfo, (*result==200)?PRG_TRACE:PRG_INFO,
-			  "Got HTTP response: %s\n", buf);
+	vpn_progress(vpninfo, (*result==200)?PRG_TRACE:PRG_INFO,
+		     _("Got HTTP response: %s\n"), buf);
 
 	/* Eat headers... */
 	while ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
 		char *colon;
 
 		if (i < 0) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Error processing HTTP response\n");
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Error processing HTTP response\n"));
 			return -EINVAL;
 		}
 		colon = strchr(buf, ':');
 		if (!colon) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Ignoring unknown HTTP response line '%s'\n", buf);
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Ignoring unknown HTTP response line '%s'\n"), buf);
 			continue;
 		}
 		*(colon++) = 0;
@@ -149,14 +153,16 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		   webvpn cookie in the verbose debug output */
 		if (!strcasecmp(buf, "Set-Cookie")) {
 			char *semicolon = strchr(colon, ';');
-			char *print_equals, *equals = strchr(colon, '=');
+			const char *print_equals;
+			char *equals = strchr(colon, '=');
 			int ret;
 
 			if (semicolon)
 				*semicolon = 0;
 
 			if (!equals) {
-				vpninfo->progress(vpninfo, PRG_ERR, "Invalid cookie offered: %s\n", buf);
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Invalid cookie offered: %s\n"), buf);
 				return -EINVAL;
 			}
 			*(equals++) = 0;
@@ -165,16 +171,23 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 			/* Don't print the webvpn cookie unless it's empty; we don't
 			   want people posting it in public with debugging output */
 			if (!strcmp(colon, "webvpn") && *equals)
-				print_equals = "<elided>";
-			vpninfo->progress(vpninfo, PRG_TRACE, "%s: %s=%s%s%s\n",
-					  buf, colon, print_equals, semicolon?";":"",
-					  semicolon?(semicolon+1):"");
+				print_equals = _("<elided>");
+			vpn_progress(vpninfo, PRG_TRACE, "%s: %s=%s%s%s\n",
+				     buf, colon, print_equals, semicolon?";":"",
+				     semicolon?(semicolon+1):"");
+
+			/* The server tends to ask for the username and password as
+			   usual, even if we've already failed because it didn't like
+			   our cert. Thankfully it does give us this hint... */
+			if (!strcmp(colon, "ClientCertAuthFailed"))
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("SSL certificate authentication failed\n"));
 
 			ret = http_add_cookie(vpninfo, colon, equals);
 			if (ret)
 				return ret;
 		} else {
-			vpninfo->progress(vpninfo, PRG_TRACE, "%s: %s\n", buf, colon);
+			vpn_progress(vpninfo, PRG_TRACE, "%s: %s\n", buf, colon);
 		}
 
 		if (!strcasecmp(buf, "Connection")) {
@@ -198,8 +211,9 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		if (!strcasecmp(buf, "Content-Length")) {
 			bodylen = atoi(colon);
 			if (bodylen < 0) {
-				vpninfo->progress(vpninfo, PRG_ERR, "Response body has negative size (%d)\n",
-						  bodylen);
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Response body has negative size (%d)\n"),
+					     bodylen);
 				return -EINVAL;
 			}
 		}
@@ -207,7 +221,9 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 			if (!strcasecmp(colon, "chunked"))
 				bodylen = BODY_CHUNKED;
 			else {
-				vpninfo->progress(vpninfo, PRG_ERR, "Unknown Transfer-Encoding: %s\n", colon);
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Unknown Transfer-Encoding: %s\n"),
+					     colon);
 				return -EINVAL;
 			}
 		}
@@ -220,10 +236,10 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		goto cont;
 
 	/* Now the body, if there is one */
-	vpninfo->progress(vpninfo, PRG_TRACE, "HTTP body %s (%d)\n", 
-			  bodylen==BODY_HTTP10?"http 1.0" :
-			  bodylen==BODY_CHUNKED?"chunked" : "length: ",
-			  bodylen);
+	vpn_progress(vpninfo, PRG_TRACE, _("HTTP body %s (%d)\n"), 
+		     bodylen==BODY_HTTP10?"http 1.0" :
+		     bodylen==BODY_CHUNKED?"chunked" : "length: ",
+		     bodylen);
 
 	/* If we were given Content-Length, it's nice and easy... */
 	if (bodylen > 0) {
@@ -233,7 +249,8 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		while (done < bodylen) {
 			i = SSL_read(vpninfo->https_ssl, body + done, bodylen - done);
 			if (i < 0) {
-				vpninfo->progress(vpninfo, PRG_ERR, "Error reading HTTP response body\n");
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Error reading HTTP response body\n"));
                                 free(body);
 				return -EINVAL;
 			}
@@ -245,7 +262,8 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 			int chunklen, lastchunk = 0;
 
 			if (i < 0) {
-				vpninfo->progress(vpninfo, PRG_ERR, "Error fetching chunk header\n");
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Error fetching chunk header\n"));
 				exit(1);
 			}
 			chunklen = strtol(buf, NULL, 16);
@@ -259,7 +277,8 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 			while (chunklen) {
 				i = SSL_read(vpninfo->https_ssl, body + done, chunklen);
 				if (i < 0) {
-					vpninfo->progress(vpninfo, PRG_ERR, "Error reading HTTP response body\n");
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Error reading HTTP response body\n"));
                                         free(body);
 					return -EINVAL;
 				}
@@ -269,10 +288,12 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		skip:
 			if ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
 				if (i < 0) {
-					vpninfo->progress(vpninfo, PRG_ERR, "Error fetching HTTP response body\n");
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Error fetching HTTP response body\n"));
 				} else {
-					vpninfo->progress(vpninfo, PRG_ERR, "Error in chunked decoding. Expected '', got: '%s'",
-							  buf);
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Error in chunked decoding. Expected '', got: '%s'"),
+						     buf);
 				}
                                 free(body);
 				return -EINVAL;
@@ -283,7 +304,8 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		}
 	} else if (bodylen == BODY_HTTP10) {
 		if (!closeconn) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Cannot receive HTTP 1.0 body without closing connection\n");
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Cannot receive HTTP 1.0 body without closing connection\n"));
 			return -EINVAL;
 		}
 
@@ -363,12 +385,13 @@ static int fetch_config(struct openconnect_info *vpninfo, char *fu, char *bu,
 		sprintf(&local_sha1_ascii[i*2], "%02x", local_sha1_bin[i]);
 
 	if (strcasecmp(server_sha1, local_sha1_ascii)) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Downloaded config file did not match intended SHA1\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Downloaded config file did not match intended SHA1\n"));
 		free(config_buf);
 		return -EINVAL;
 	}
 
-	result = vpninfo->write_new_config(vpninfo, config_buf, buflen);
+	result = vpninfo->write_new_config(vpninfo->cbdata, config_buf, buflen);
 	free(config_buf);
 	return result;
 }
@@ -379,30 +402,32 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 	int fd, ret;
 
 	if (!vpninfo->uid_csd_given && !vpninfo->csd_wrapper) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Error: Server asked us to download and run a 'Cisco Secure Desktop' trojan.\n"
-				  "This facility is disabled by default for security reasons, so you may wish to enable it.");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Error: Server asked us to download and run a 'Cisco Secure Desktop' trojan.\n"
+			       "This facility is disabled by default for security reasons, so you may wish to enable it."));
 		return -EPERM;
 	}
 
 #ifndef __linux__
-	vpninfo->progress(vpninfo, PRG_INFO,
-			  "Trying to run Linux CSD trojan script.");
+	vpn_progress(vpninfo, PRG_INFO,
+		     _("Trying to run Linux CSD trojan script."));
 #endif
 
 	sprintf(fname, "/tmp/csdXXXXXX");
 	fd = mkstemp(fname);
 	if (fd < 0) {
 		int err = -errno;
-		vpninfo->progress(vpninfo, PRG_ERR, "Failed to open temporary CSD script file: %s\n",
-				  strerror(errno));
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to open temporary CSD script file: %s\n"),
+			     strerror(errno));
 		return err;
 	}
 
 	ret = proxy_write(fd, (void *)buf, buflen);
 	if (ret) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Failed to write temporary CSD script file: %s\n",
-				  strerror(ret));
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to write temporary CSD script file: %s\n"),
+			     strerror(ret));
 		return ret;
 	}
 	fchmod(fd, 0755);
@@ -420,26 +445,26 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 			struct passwd *pw;
 
 			if (setuid(vpninfo->uid_csd)) {
-				fprintf(stderr, "Failed to set uid %d\n",
+				fprintf(stderr, _("Failed to set uid %d\n"),
 					vpninfo->uid_csd);
 				exit(1);
 			}
 			if (!(pw = getpwuid(vpninfo->uid_csd))) {
-				fprintf(stderr, "Invalid user uid=%d\n",
+				fprintf(stderr, _("Invalid user uid=%d\n"),
 					vpninfo->uid_csd);
 				exit(1);
 			}
 			setenv("HOME", pw->pw_dir, 1);
 			if (chdir(pw->pw_dir)) {
-				fprintf(stderr, "Failed to change to CSD home directory '%s': %s\n",
+				fprintf(stderr, _("Failed to change to CSD home directory '%s': %s\n"),
 					pw->pw_dir, strerror(errno));
 				exit(1);
 			}
 		}
 		if (vpninfo->uid_csd == 0 && !vpninfo->csd_wrapper) {
-			fprintf(stderr, "Warning: you are running insecure "
-				"CSD code with root privileges\n"
-				"\t Use command line option \"--csd-user\"\n");
+			fprintf(stderr, _("Warning: you are running insecure "
+					  "CSD code with root privileges\n"
+					  "\t Use command line option \"--csd-user\"\n"));
 		}
 		if (vpninfo->uid_csd_given == 2) {	       
 			/* The NM tool really needs not to get spurious output
@@ -449,12 +474,12 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 		if (vpninfo->csd_wrapper)
 			csd_argv[i++] = vpninfo->csd_wrapper;
 		csd_argv[i++] = fname;
-		csd_argv[i++] = "-ticket";
+		csd_argv[i++]= (char *)"-ticket";
 		if (asprintf(&csd_argv[i++], "\"%s\"", vpninfo->csd_ticket) == -1)
 			return -ENOMEM;
-		csd_argv[i++] = "-stub";
-		csd_argv[i++] = "\"0\"";
-		csd_argv[i++] = "-group";
+		csd_argv[i++]= (char *)"-stub";
+		csd_argv[i++]= (char *)"\"0\"";
+		csd_argv[i++]= (char *)"-group";
 		if (asprintf(&csd_argv[i++], "\"%s\"", vpninfo->authgroup?:"") == -1)
 			return -ENOMEM;
 
@@ -464,26 +489,27 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 		else
 			ccertbuf[0] = 0;
 
-		csd_argv[i++] = "-certhash";
+		csd_argv[i++]= (char *)"-certhash";
 		if (asprintf(&csd_argv[i++], "\"%s:%s\"", scertbuf, ccertbuf) == -1)
 			return -ENOMEM;
-		csd_argv[i++] = "-url";
+		csd_argv[i++]= (char *)"-url";
 		if (asprintf(&csd_argv[i++], "\"https://%s%s\"", vpninfo->hostname, vpninfo->csd_starturl) == -1)
 			return -ENOMEM;
 		/* WTF would it want to know this for? */
-		csd_argv[i++] = "-vpnclient";
-		csd_argv[i++] = "\"/opt/cisco/vpn/bin/vpnui";
-		csd_argv[i++] = "-connect";
+		csd_argv[i++]= (char *)"-vpnclient";
+		csd_argv[i++]= (char *)"\"/opt/cisco/vpn/bin/vpnui";
+		csd_argv[i++]= (char *)"-connect";
 		if (asprintf(&csd_argv[i++], "https://%s/%s", vpninfo->hostname, vpninfo->csd_preurl) == -1)
 			return -ENOMEM;
-		csd_argv[i++] = "-connectparam";
+		csd_argv[i++]= (char *)"-connectparam";
 		if (asprintf(&csd_argv[i++], "#csdtoken=%s\"", vpninfo->csd_token) == -1)
 			return -ENOMEM;
-		csd_argv[i++] = "-langselen";
+		csd_argv[i++]= (char *)"-langselen";
 		csd_argv[i++] = NULL;
 
 		execv(csd_argv[0], csd_argv);
-		vpninfo->progress(vpninfo, PRG_ERR, "Failed to exec CSD script %s\n", csd_argv[0]);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to exec CSD script %s\n"), csd_argv[0]);
 		exit(1);
 	}
 
@@ -500,7 +526,7 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 }
 
 #ifdef __sun__
-char *local_strcasestr(const char *haystack, const char *needle)
+static char *local_strcasestr(const char *haystack, const char *needle)
 {
 	int hlen = strlen(haystack);
 	int nlen = strlen(needle);
@@ -595,8 +621,8 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 	char *form_buf = NULL;
 	int result, buflen;
 	char request_body[2048];
-	char *request_body_type = NULL;
-	char *method = "GET";
+	const char *request_body_type = NULL;
+	const char *method = "GET";
 
  retry:
 	if (form_buf) {
@@ -604,8 +630,9 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 		form_buf = NULL;
 	}
 	if (!vpninfo->https_ssl && openconnect_open_https(vpninfo)) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Failed to open HTTPS connection to %s\n",
-			vpninfo->hostname);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to open HTTPS connection to %s\n"),
+			     vpninfo->hostname);
 		return -EINVAL;
 	}
 
@@ -641,13 +668,13 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 		sprintf(buf + strlen(buf), "%s", request_body);
 
 	if (vpninfo->port == 443)
-		vpninfo->progress(vpninfo, PRG_INFO, "%s https://%s/%s\n",
-				  method, vpninfo->hostname,
-				  vpninfo->urlpath ?: "");
+		vpn_progress(vpninfo, PRG_INFO, "%s https://%s/%s\n",
+			     method, vpninfo->hostname,
+			     vpninfo->urlpath ?: "");
 	else
-		vpninfo->progress(vpninfo, PRG_INFO, "%s https://%s:%d/%s\n",
-				  method, vpninfo->hostname, vpninfo->port,
-				  vpninfo->urlpath ?: "");
+		vpn_progress(vpninfo, PRG_INFO, "%s https://%s:%d/%s\n",
+			     method, vpninfo->hostname, vpninfo->port,
+			     vpninfo->urlpath ?: "");
 
 	SSL_write(vpninfo->https_ssl, buf, strlen(buf));
 
@@ -670,8 +697,9 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 
 			ret = internal_parse_url(vpninfo->redirect_url, NULL, &host, &port, &vpninfo->urlpath, 0);
 			if (ret) {
-				vpninfo->progress(vpninfo, PRG_ERR, "Failed to parse redirected URL '%s': %s\n",
-						  vpninfo->redirect_url, strerror(-ret));
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Failed to parse redirected URL '%s': %s\n"),
+					     vpninfo->redirect_url, strerror(-ret));
 				free(vpninfo->redirect_url);
 				free(form_buf);
 				return ret;
@@ -729,9 +757,9 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 				if (asprintf(&vpninfo->urlpath, "%s/%s",
 					     oldurl, vpninfo->redirect_url) == -1) {
 					int err = -errno;
-					vpninfo->progress(vpninfo, PRG_ERR,
-							  "Allocating new path for relative redirect failed: %s\n",
-							  strerror(-err));
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Allocating new path for relative redirect failed: %s\n"),
+						     strerror(-err));
 					return err;
 				}
 				free(oldurl);
@@ -742,9 +770,9 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 		}
 	}
 	if (!form_buf || result != 200) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Unexpected %d result from server\n",
-				  result);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Unexpected %d result from server\n"),
+			     result);
 		free(form_buf);
 		return -EINVAL;
 	}
@@ -762,12 +790,14 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 	if (strncmp(form_buf, "<?xml", 5)) {
 		/* Not XML? Perhaps it's HTML with a refresh... */
 		if (strcasestr(form_buf, "http-equiv=\"refresh\"")) {
-			vpninfo->progress(vpninfo, PRG_INFO, "Refreshing %s after 1 second...\n",
-					  vpninfo->urlpath);
+			vpn_progress(vpninfo, PRG_INFO,
+				     _("Refreshing %s after 1 second...\n"),
+				     vpninfo->urlpath);
 			sleep(1);
 			goto retry;
 		}
-		vpninfo->progress(vpninfo, PRG_ERR, "Unknown response from server\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Unknown response from server\n"));
 		free(form_buf);
 		return -EINVAL;
 	}
@@ -822,7 +852,7 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 	return 0;
 }
 
-char *openconnect_create_useragent(char *base)
+char *openconnect_create_useragent(const char *base)
 {
 	char *uagent;
 
@@ -892,15 +922,15 @@ static int proxy_read(int fd, unsigned char *buf, size_t len)
 }
 
 static const char *socks_errors[] = {
-	"request granted",
-	"general failure",
-	"connection not allowed by ruleset",
-	"network unreachable",
-	"host unreachable",
-	"connection refused by destination host",
-	"TTL expired",
-	"command not supported / protocol error",
-	"address type not supported"
+	N_("request granted"),
+	N_("general failure"),
+	N_("connection not allowed by ruleset"),
+	N_("network unreachable"),
+	N_("host unreachable"),
+	N_("connection refused by destination host"),
+	N_("TTL expired"),
+	N_("command not supported / protocol error"),
+	N_("address type not supported")
 };
 
 static int process_socks_proxy(struct openconnect_info *vpninfo, int ssl_sock)
@@ -913,39 +943,40 @@ static int process_socks_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 	buf[2] = 0; /* No auth supported */
 
 	if ((i = proxy_write(ssl_sock, buf, 3))) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Error writing auth request to SOCKS proxy: %s\n",
-				  strerror(-i));
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Error writing auth request to SOCKS proxy: %s\n"),
+			     strerror(-i));
 		return i;
 	}
 	
 	if ((i = proxy_read(ssl_sock, buf, 2))) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Error reading auth response from SOCKS proxy: %s\n",
-				  strerror(-i));
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Error reading auth response from SOCKS proxy: %s\n"),
+			     strerror(-i));
 		return i;
 	}
 	if (buf[0] != 5) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Unexpected auth response from SOCKS proxy: %02x %02x\n",
-				  buf[0], buf[1]);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Unexpected auth response from SOCKS proxy: %02x %02x\n"),
+			     buf[0], buf[1]);
 		return -EIO;
 	}
 	if (buf[1]) {
 	socks_err:
 		if (buf[1] < sizeof(socks_errors) / sizeof(socks_errors[0]))
-			vpninfo->progress(vpninfo, PRG_ERR,
-					  "SOCKS proxy error %02x: %s\n",
-					  buf[1], socks_errors[buf[1]]);
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("SOCKS proxy error %02x: %s\n"),
+				     buf[1], _(socks_errors[buf[1]]));
 		else
-			vpninfo->progress(vpninfo, PRG_ERR,
-					  "SOCKS proxy error %02x\n",
-					  buf[1]);
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("SOCKS proxy error %02x\n"),
+				     buf[1]);
 		return -EIO;
 	}
 
-	vpninfo->progress(vpninfo, PRG_INFO, "Requesting SOCKS proxy connection to %s:%d\n",
-			  vpninfo->hostname, vpninfo->port);
+	vpn_progress(vpninfo, PRG_INFO,
+		     _("Requesting SOCKS proxy connection to %s:%d\n"),
+		     vpninfo->hostname, vpninfo->port);
 
 	buf[0] = 5; /* SOCKS version */
 	buf[1] = 1; /* CONNECT */
@@ -958,23 +989,23 @@ static int process_socks_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 	buf[i++] = vpninfo->port & 0xff;
 
 	if ((i = proxy_write(ssl_sock, buf, i))) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Error writing connect request to SOCKS proxy: %s\n",
-				  strerror(-i));
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Error writing connect request to SOCKS proxy: %s\n"),
+			     strerror(-i));
 		return i;
 	}
 	/* Read 5 bytes -- up to and including the first byte of the returned
 	   address (which might be the length byte of a domain name) */
 	if ((i = proxy_read(ssl_sock, buf, 5))) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Error reading connect response from SOCKS proxy: %s\n",
-				  strerror(-i));
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Error reading connect response from SOCKS proxy: %s\n"),
+			     strerror(-i));
 		return i;
 	}
 	if (buf[0] != 5) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Unexpected connect response from SOCKS proxy: %02x %02x...\n",
-				  buf[0], buf[1]);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Unexpected connect response from SOCKS proxy: %02x %02x...\n"),
+			     buf[0], buf[1]);
 		return -EIO;
 	}
 	if (buf[1])
@@ -992,16 +1023,16 @@ static int process_socks_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 		i = 17;
 		break;
 	default:
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Unexpected address type %02x in SOCKS connect response\n",
-				  buf[3]);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Unexpected address type %02x in SOCKS connect response\n"),
+			     buf[3]);
 		return -EIO;
 	}
 
 	if ((i = proxy_read(ssl_sock, buf, i))) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Error reading connect response from SOCKS proxy: %s\n",
-				  strerror(-i));
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Error reading connect response from SOCKS proxy: %s\n"),
+			     strerror(-i));
 		return i;
 	}
 	return 0;
@@ -1020,42 +1051,46 @@ static int process_http_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 	sprintf(buf + strlen(buf), "Accept-Encoding: identity\r\n");
 	sprintf(buf + strlen(buf), "\r\n");
 
-	vpninfo->progress(vpninfo, PRG_INFO, "Requesting HTTP proxy connection to %s:%d\n",
-			  vpninfo->hostname, vpninfo->port);
+	vpn_progress(vpninfo, PRG_INFO,
+		     _("Requesting HTTP proxy connection to %s:%d\n"),
+		     vpninfo->hostname, vpninfo->port);
 
 	if (proxy_write(ssl_sock, (unsigned char *)buf, strlen(buf))) {
 		result = -errno;
-		vpninfo->progress(vpninfo, PRG_ERR, "Sending proxy request failed: %s\n",
-				  strerror(errno));
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Sending proxy request failed: %s\n"),
+			     strerror(errno));
 		return result;
 	}
 
 	if (proxy_gets(ssl_sock, buf, sizeof(buf)) < 0) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Error fetching proxy response\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Error fetching proxy response\n"));
 		return -EIO;
 	}
 
 	if (strncmp(buf, "HTTP/1.", 7) || (buf[7] != '0' && buf[7] != '1') ||
 	    buf[8] != ' ' || !(result = atoi(buf+9))) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Failed to parse proxy response '%s'\n",
-				  buf);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to parse proxy response '%s'\n"), buf);
 		return -EINVAL;
 	}
 
 	if (result != 200) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Proxy CONNECT request failed: %s\n",
-				  buf);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Proxy CONNECT request failed: %s\n"), buf);
 		return -EIO;
 	}
 
 	while ((buflen = proxy_gets(ssl_sock, buf, sizeof(buf)))) {
 		if (buflen < 0) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Failed to read proxy response\n");
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to read proxy response\n"));
 			return -EIO;
 		}
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Unexpected continuation line after CONNECT response: '%s'\n",
-				  buf);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Unexpected continuation line after CONNECT response: '%s'\n"),
+			     buf);
 	}
 
 	return 0;
@@ -1070,8 +1105,8 @@ int process_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 	    !strcmp(vpninfo->proxy_type, "socks5"))
 		return process_socks_proxy(vpninfo, ssl_sock);
 
-	vpninfo->progress(vpninfo, PRG_ERR, "Unknown proxy type '%s'\n",
-				  vpninfo->proxy_type);
+	vpn_progress(vpninfo, PRG_ERR, _("Unknown proxy type '%s'\n"),
+		     vpninfo->proxy_type);
 	return -EIO;
 }
 
@@ -1097,8 +1132,8 @@ int openconnect_set_http_proxy(struct openconnect_info *vpninfo, char *proxy)
 	    strcmp(vpninfo->proxy_type, "http") &&
 	    strcmp(vpninfo->proxy_type, "socks") &&
 	    strcmp(vpninfo->proxy_type, "socks5")) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Only http or socks(5) proxies supported\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Only http or socks(5) proxies supported\n"));
 		free(vpninfo->proxy_type);
 		vpninfo->proxy_type = NULL;
 		free(vpninfo->proxy);
