@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2010 Intel Corporation.
+ * Copyright © 2008-2011 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
@@ -24,37 +24,53 @@
  */
 
 #include <stdio.h>
+#ifdef ANDROID
+#include <android/log.h>
+#else
 #include <syslog.h>
+#endif
 #include <stdarg.h>
 #include <stdlib.h>
 #include <signal.h>
 #include <string.h>
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
+#endif
 #include <errno.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <pwd.h>
-#include <sys/syslog.h>
 #include <sys/utsname.h>
 #include <sys/types.h>
 #include <openssl/ui.h>
-#ifdef OPENCONNECT_LIBPROXY
+#ifdef LIBPROXY_HDR
 #include LIBPROXY_HDR
 #endif
 
-#define _GNU_SOURCE
 #include <getopt.h>
 
 #include "openconnect-internal.h"
 
-static int write_new_config(struct openconnect_info *vpninfo, char *buf, int buflen);
-static void write_progress(struct openconnect_info *info, int level, const char *fmt, ...);
-static void syslog_progress(struct openconnect_info *info, int level, const char *fmt, ...);
-static int validate_peer_cert(struct openconnect_info *info, X509 *peer_cert, const char *reason);
+static int write_new_config(void *_vpninfo,
+			    char *buf, int buflen);
+static void write_progress(void *_vpninfo,
+			   int level, const char *fmt, ...);
+static void syslog_progress(void *_vpninfo,
+			    int level, const char *fmt, ...);
+static int validate_peer_cert(void *_vpninfo,
+			      X509 *peer_cert, const char *reason);
+
+/* A sanity check that the openconnect executable is running against a
+   library of the same version */
+#define openconnect_version openconnect_binary_version
+#include "version.c"
+#undef openconnect_version
 
 int verbose = PRG_INFO;
 int background;
 int do_passphrase_from_fsid;
 int nocertcheck;
+int non_inter;
 
 enum {
 	OPT_AUTHGROUP = 0x100,
@@ -73,20 +89,24 @@ enum {
 	OPT_NO_HTTP_KEEPALIVE,
 	OPT_NO_PASSWD,
 	OPT_NO_PROXY,
+	OPT_PIDFILE,
 	OPT_PASSWORD_ON_STDIN,
 	OPT_PRINTCOOKIE,
 	OPT_RECONNECT_TIMEOUT,
 	OPT_SERVERCERT,
 	OPT_USERAGENT,
+	OPT_NON_INTER,
 };
 
 static struct option long_options[] = {
 	{"background", 0, 0, 'b'},
+	{"pid-file", 1, 0, OPT_PIDFILE},
 	{"certificate", 1, 0, 'c'},
 	{"sslkey", 1, 0, 'k'},
 	{"cookie", 1, 0, 'C'},
 	{"deflate", 0, 0, 'd'},
 	{"no-deflate", 0, 0, 'D'},
+	{"cert-expire-warning", 1, 0, 'e'},
 	{"usergroup", 1, 0, 'g'},
 	{"help", 0, 0, 'h'},
 	{"interface", 1, 0, 'i'},
@@ -125,60 +145,64 @@ static struct option long_options[] = {
 	{"no-http-keepalive", 0, 0, OPT_NO_HTTP_KEEPALIVE},
 	{"no-cert-check", 0, 0, OPT_NO_CERT_CHECK},
 	{"force-dpd", 1, 0, OPT_FORCE_DPD},
+	{"non-inter", 0, 0, OPT_NON_INTER},
 	{NULL, 0, 0, 0},
 };
 
-void usage(void)
+static void usage(void)
 {
-	printf("Usage:  openconnect [options] <server>\n");
-	printf("Open client for Cisco AnyConnect VPN, version %s\n\n", openconnect_version);
-	printf("  -b, --background                Continue in background after startup\n");
-	printf("  -c, --certificate=CERT          Use SSL client certificate CERT\n");
-	printf("  -k, --sslkey=KEY                Use SSL private key file KEY\n");
-	printf("  -K, --key-type=TYPE             Private key type (PKCS#12 / TPM / PEM)\n");
-	printf("  -C, --cookie=COOKIE             Use WebVPN cookie COOKIE\n");
-	printf("      --cookie-on-stdin           Read cookie from standard input\n");
-	printf("  -d, --deflate                   Enable compression (default)\n");
-	printf("  -D, --no-deflate                Disable compression\n");
-	printf("      --force-dpd=INTERVAL        Set minimum Dead Peer Detection interval\n");
-	printf("  -g, --usergroup=GROUP           Set login usergroup\n");
-	printf("  -h, --help                      Display help text\n");
-	printf("  -i, --interface=IFNAME          Use IFNAME for tunnel interface\n");
-	printf("  -l, --syslog                    Use syslog for progress messages\n");
-	printf("  -U, --setuid=USER               Drop privileges after connecting\n");
-	printf("      --csd-user=USER             Drop privileges during CSD execution\n");
-	printf("      --csd-wrapper=SCRIPT        Run SCRIPT instead of CSD binary\n");
-	printf("  -m, --mtu=MTU                   Request MTU from server\n");
-	printf("  -p, --key-password=PASS         Set key passphrase or TPM SRK PIN\n");
-	printf("      --key-password-from-fsid    Key passphrase is fsid of file system\n");
-	printf("  -P, --proxy=URL                 Set proxy server\n");
-	printf("      --no-proxy                  Disable proxy\n");
-	printf("      --libproxy                  Use libproxy to automatically configure proxy\n");
-#ifndef OPENCONNECT_LIBPROXY
-	printf("                                  (NOTE: libproxy disabled in this build)\n");
+	printf(_("Usage:  openconnect [options] <server>\n"));
+	printf(_("Open client for Cisco AnyConnect VPN, version %s\n\n"), openconnect_version);
+	printf("  -b, --background                %s\n", _("Continue in background after startup"));
+	printf("      --pid-file=PIDFILE          %s\n", _("Write the daemons pid to this file"));
+	printf("  -c, --certificate=CERT          %s\n", _("Use SSL client certificate CERT"));
+	printf("  -e, --cert-expire-warning=DAYS  %s\n", _("Warn when certificate lifetime < DAYS"));
+	printf("  -k, --sslkey=KEY                %s\n", _("Use SSL private key file KEY"));
+	printf("  -K, --key-type=TYPE             %s\n", _("Private key type (PKCS#12 / TPM / PEM)"));
+	printf("  -C, --cookie=COOKIE             %s\n", _("Use WebVPN cookie COOKIE"));
+	printf("      --cookie-on-stdin           %s\n", _("Read cookie from standard input"));
+	printf("  -d, --deflate                   %s\n", _("Enable compression (default)"));
+	printf("  -D, --no-deflate                %s\n", _("Disable compression"));
+	printf("      --force-dpd=INTERVAL        %s\n", _("Set minimum Dead Peer Detection interval"));
+	printf("  -g, --usergroup=GROUP           %s\n", _("Set login usergroup"));
+	printf("  -h, --help                      %s\n", _("Display help text"));
+	printf("  -i, --interface=IFNAME          %s\n", _("Use IFNAME for tunnel interface"));
+	printf("  -l, --syslog                    %s\n", _("Use syslog for progress messages"));
+	printf("  -U, --setuid=USER               %s\n", _("Drop privileges after connecting"));
+	printf("      --csd-user=USER             %s\n", _("Drop privileges during CSD execution"));
+	printf("      --csd-wrapper=SCRIPT        %s\n", _("Run SCRIPT instead of CSD binary"));
+	printf("  -m, --mtu=MTU                   %s\n", _("Request MTU from server"));
+	printf("  -p, --key-password=PASS         %s\n", _("Set key passphrase or TPM SRK PIN"));
+	printf("      --key-password-from-fsid    %s\n", _("Key passphrase is fsid of file system"));
+	printf("  -P, --proxy=URL                 %s\n", _("Set proxy server"));
+	printf("      --no-proxy                  %s\n", _("Disable proxy"));
+	printf("      --libproxy                  %s\n", _("Use libproxy to automatically configure proxy"));
+#ifndef LIBPROXY_HDR
+	printf("                                  %s\n", _("(NOTE: libproxy disabled in this build)"));
 #endif
-	printf("  -q, --quiet                     Less output\n");
-	printf("  -Q, --queue-len=LEN             Set packet queue limit to LEN pkts\n");
-	printf("  -s, --script=SCRIPT             Use vpnc-compatible config script\n");
-	printf("  -S, --script-tun                Pass traffic to 'script' program, not tun\n");
-	printf("  -u, --user=NAME                 Set login username\n");
-	printf("  -V, --version                   Report version number\n");
-	printf("  -v, --verbose                   More output\n");
-	printf("  -x, --xmlconfig=CONFIG          XML config file\n");
-	printf("      --authgroup=GROUP           Choose authentication login selection\n");
-	printf("      --cookieonly                Fetch webvpn cookie only; don't connect\n");
-	printf("      --printcookie               Print webvpn cookie before connecting\n");
-	printf("      --cafile=FILE               Cert file for server verification\n");
-	printf("      --disable-ipv6              Do not ask for IPv6 connectivity\n");
-	printf("      --dtls-ciphers=LIST         OpenSSL ciphers to support for DTLS\n");
-	printf("      --no-dtls                   Disable DTLS\n");
-	printf("      --no-http-keepalive         Disable HTTP connection re-use\n");
-	printf("      --no-passwd                 Disable password/SecurID authentication\n");
-	printf("      --no-cert-check             Do not require server SSL cert to be valid\n");
-	printf("      --passwd-on-stdin           Read password from standard input\n");
-	printf("      --reconnect-timeout         Connection retry timeout in seconds\n");
-	printf("      --servercert=FINGERPRINT    Server's certificate SHA1 fingerprint\n");
-	printf("      --useragent=STRING          HTTP header User-Agent: field\n");
+	printf("  -q, --quiet                     %s\n", _("Less output"));
+	printf("  -Q, --queue-len=LEN             %s\n", _("Set packet queue limit to LEN pkts"));
+	printf("  -s, --script=SCRIPT             %s\n", _("Shell command line for using a vpnc-compatible config script"));
+	printf("  -S, --script-tun                %s\n", _("Pass traffic to 'script' program, not tun"));
+	printf("  -u, --user=NAME                 %s\n", _("Set login username"));
+	printf("  -V, --version                   %s\n", _("Report version number"));
+	printf("  -v, --verbose                   %s\n", _("More output"));
+	printf("  -x, --xmlconfig=CONFIG          %s\n", _("XML config file"));
+	printf("      --authgroup=GROUP           %s\n", _("Choose authentication login selection"));
+	printf("      --cookieonly                %s\n", _("Fetch webvpn cookie only; don't connect"));
+	printf("      --printcookie               %s\n", _("Print webvpn cookie before connecting"));
+	printf("      --cafile=FILE               %s\n", _("Cert file for server verification"));
+	printf("      --disable-ipv6              %s\n", _("Do not ask for IPv6 connectivity"));
+	printf("      --dtls-ciphers=LIST         %s\n", _("OpenSSL ciphers to support for DTLS"));
+	printf("      --no-dtls                   %s\n", _("Disable DTLS"));
+	printf("      --no-http-keepalive         %s\n", _("Disable HTTP connection re-use"));
+	printf("      --no-passwd                 %s\n", _("Disable password/SecurID authentication"));
+	printf("      --no-cert-check             %s\n", _("Do not require server SSL cert to be valid"));
+	printf("      --non-inter                 %s\n", _("Do not expect user input; exit if it is required"));
+	printf("      --passwd-on-stdin           %s\n", _("Read password from standard input"));
+	printf("      --reconnect-timeout         %s\n", _("Connection retry timeout in seconds"));
+	printf("      --servercert=FINGERPRINT    %s\n", _("Server's certificate SHA1 fingerprint"));
+	printf("      --useragent=STRING          %s\n", _("HTTP header User-Agent: field"));
 	exit(1);
 }
 
@@ -186,11 +210,11 @@ static void read_stdin(char **string)
 {
 	char *c = malloc(100);
 	if (!c) {
-		fprintf(stderr, "Allocation failure for string from stdin\n");
+		fprintf(stderr, _("Allocation failure for string from stdin\n"));
 		exit(1);
 	}
 	if (!fgets(c, 100, stdin)) {
-		perror("fgets (stdin)");
+		perror(_("fgets (stdin)"));
 		exit(1);
 	}
 
@@ -219,12 +243,25 @@ int main(int argc, char **argv)
 	int autoproxy = 0;
 	uid_t uid = getuid();
 	int opt;
+	char *pidfile = NULL;
+	FILE *fp = NULL;
 
+#ifdef ENABLE_NLS
+	bindtextdomain("openconnect", LOCALEDIR);
+	setlocale(LC_ALL, "");
+#endif
+
+	if (strcmp(openconnect_version, openconnect_binary_version)) {
+		fprintf(stderr, _("WARNING: This version of openconnect is %s but\n"
+				  "         the libopenconnect library is %s\n"),
+			openconnect_binary_version, openconnect_version);
+	}
+			
 	openconnect_init_openssl();
 
 	vpninfo = malloc(sizeof(*vpninfo));
 	if (!vpninfo) {
-		fprintf(stderr, "Failed to allocate vpninfo structure\n");
+		fprintf(stderr, _("Failed to allocate vpninfo structure\n"));
 		exit(1);
 	}
 	memset(vpninfo, 0, sizeof(*vpninfo));
@@ -241,13 +278,15 @@ int main(int argc, char **argv)
 	vpninfo->uid_csd = 0;
 	vpninfo->uid_csd_given = 0;
 	vpninfo->validate_peer_cert = validate_peer_cert;
+	vpninfo->cbdata = vpninfo;
+	vpninfo->cert_expire_warning = 60 * 86400;
 
 	if (!uname(&utsbuf))
 		vpninfo->localname = utsbuf.nodename;
 	else
 		vpninfo->localname = "localhost";
 
-	while ((opt = getopt_long(argc, argv, "bC:c:Ddg:hi:k:K:lpP:Q:qSs:U:u:Vvx:",
+	while ((opt = getopt_long(argc, argv, "bC:c:e:Ddg:hi:k:K:lpP:Q:qSs:U:u:Vvx:",
 				  long_options, NULL))) {
 		if (opt < 0)
 			break;
@@ -256,6 +295,9 @@ int main(int argc, char **argv)
 		case OPT_CAFILE:
 			vpninfo->cafile = optarg;
 			break;
+		case OPT_PIDFILE:
+			pidfile = optarg;
+			break;
 		case OPT_SERVERCERT:
 			vpninfo->servercert = optarg;
 			break;
@@ -281,6 +323,10 @@ int main(int argc, char **argv)
 		case OPT_NO_PASSWD:
 			vpninfo->nopasswd = 1;
 			break;
+		case OPT_NON_INTER:
+			non_inter = 1;
+			vpninfo->nopasswd = 1;
+			break;
 		case OPT_RECONNECT_TIMEOUT:
 			vpninfo->reconnect_timeout = atoi(optarg);
 			break;
@@ -299,6 +345,9 @@ int main(int argc, char **argv)
 		case 'c':
 			vpninfo->cert = optarg;
 			break;
+		case 'e':
+			vpninfo->cert_expire_warning = 86400 * atoi(optarg);
+			break;
 		case 'k':
 			vpninfo->sslkey = optarg;
 			break;
@@ -311,7 +360,7 @@ int main(int argc, char **argv)
 			} else if (!strcasecmp(optarg, "PEM")) {
 				vpninfo->cert_type = CERT_TYPE_PEM;
 			} else {
-				fprintf(stderr, "Unknown certificate type '%s'\n",
+				fprintf(stderr, _("Unknown certificate type '%s'\n"),
 					optarg);
 				usage();
 			}
@@ -336,7 +385,7 @@ int main(int argc, char **argv)
 		case 'm':
 			vpninfo->mtu = atol(optarg);
 			if (vpninfo->mtu < 576) {
-				fprintf(stderr, "MTU %d too small\n", vpninfo->mtu);
+				fprintf(stderr, _("MTU %d too small\n"), vpninfo->mtu);
 				vpninfo->mtu = 576;
 			}
 			break;
@@ -351,16 +400,13 @@ int main(int argc, char **argv)
 			autoproxy = 0;
 			proxy = NULL;
 		case OPT_LIBPROXY:
-#ifndef OPENCONNECT_LIBPROXY
-			fprintf(stderr, "This version of openconnect was built without libproxy support\n");
-			exit(1);
-#endif
 			autoproxy = 1;
 			proxy = NULL;
 			break;
 		case OPT_NO_HTTP_KEEPALIVE:
-			fprintf(stderr, "Disabling all HTTP connection re-use due to --no-http-keepalive option.\n"
-				"If this helps, please report to <openconnect-devel@lists.infradead.org>.\n");
+			fprintf(stderr,
+				_("Disabling all HTTP connection re-use due to --no-http-keepalive option.\n"
+				  "If this helps, please report to <openconnect-devel@lists.infradead.org>.\n"));
 			vpninfo->no_http_keepalive = 1;
 			break;
 		case OPT_NO_CERT_CHECK:
@@ -381,7 +427,7 @@ int main(int argc, char **argv)
 			if (strend[0]) {
 				struct passwd *pw = getpwnam(optarg);
 				if (!pw) {
-					fprintf(stderr, "Invalid user \"%s\"\n",
+					fprintf(stderr, _("Invalid user \"%s\"\n"),
 						optarg);
 					exit(1);
 				}
@@ -395,7 +441,7 @@ int main(int argc, char **argv)
 			if (strend[0]) {
 				struct passwd *pw = getpwnam(optarg);
 				if (!pw) {
-					fprintf(stderr, "Invalid user \"%s\"\n",
+					fprintf(stderr, _("Invalid user \"%s\"\n"),
 						optarg);
 					exit(1);
 				}
@@ -413,7 +459,7 @@ int main(int argc, char **argv)
 		case 'Q':
 			vpninfo->max_qlen = atol(optarg);
 			if (!vpninfo->max_qlen) {
-				fprintf(stderr, "Queue length zero not permitted; using 1\n");
+				fprintf(stderr, _("Queue length zero not permitted; using 1\n"));
 				vpninfo->max_qlen = 1;
 			}
 			break;
@@ -424,7 +470,7 @@ int main(int argc, char **argv)
 			verbose = PRG_TRACE;
 			break;
 		case 'V':
-			printf("OpenConnect version %s\n", openconnect_version);
+			printf(_("OpenConnect version %s\n"), openconnect_version);
 			exit(0);
 		case 'x':
 			vpninfo->xmlconfig = optarg;
@@ -446,7 +492,7 @@ int main(int argc, char **argv)
 	}
 
 	if (optind != argc - 1) {
-		fprintf(stderr, "No server specified\n");
+		fprintf(stderr, _("No server specified\n"));
 		usage();
 	}
 
@@ -455,15 +501,22 @@ int main(int argc, char **argv)
 
 	vpninfo->progress = write_progress;
 
-#ifdef OPENCONNECT_LIBPROXY
-	if (autoproxy)
+	if (autoproxy) {
+#ifdef LIBPROXY_HDR
 		vpninfo->proxy_factory = px_proxy_factory_new();
+#else
+		fprintf(stderr, _("This version of openconnect was built without libproxy support\n"));
+		exit(1);
 #endif
+	}
+
 	if (proxy && openconnect_set_http_proxy(vpninfo, strdup(proxy)))
 		exit(1);
 
 	if (use_syslog) {
+#ifndef ANDROID
 		openlog("openconnect", LOG_PID, LOG_DAEMON);
+#endif
 		vpninfo->progress = syslog_progress;
 	}
 
@@ -486,12 +539,12 @@ int main(int argc, char **argv)
 
 		if (internal_parse_url(url, &scheme, &vpninfo->hostname, &vpninfo->port,
 			      &group, 443)) {
-			fprintf(stderr, "Failed to parse server URL '%s'\n",
+			fprintf(stderr, _("Failed to parse server URL '%s'\n"),
 				url);
 			exit(1);
 		}
 		if (scheme && strcmp(scheme, "https")) {
-			fprintf(stderr, "Only https:// permitted for server URL\n");
+			fprintf(stderr, _("Only https:// permitted for server URL\n"));
 			exit(1);
 		}
 		if (group) {
@@ -507,7 +560,7 @@ int main(int argc, char **argv)
 #endif
 
 	if (!vpninfo->cookie && openconnect_obtain_cookie(vpninfo)) {
-		fprintf(stderr, "Failed to obtain WebVPN cookie\n");
+		fprintf(stderr, _("Failed to obtain WebVPN cookie\n"));
 		exit(1);
 	}
 
@@ -518,60 +571,81 @@ int main(int argc, char **argv)
 			exit(0);
 	}
 	if (make_cstp_connection(vpninfo)) {
-		fprintf(stderr, "Creating SSL connection failed\n");
+		fprintf(stderr, _("Creating SSL connection failed\n"));
 		exit(1);
 	}
 
 	if (setup_tun(vpninfo)) {
-		fprintf(stderr, "Set up tun device failed\n");
+		fprintf(stderr, _("Set up tun device failed\n"));
 		exit(1);
 	}
 
 	if (uid != getuid()) {
 		if (setuid(uid)) {
-			fprintf(stderr, "Failed to set uid %d\n", uid);
+			fprintf(stderr, _("Failed to set uid %d\n"), uid);
 			exit(1);
 		}
 	}
 
 	if (vpninfo->dtls_attempt_period && setup_dtls(vpninfo))
-		fprintf(stderr, "Set up DTLS failed; using SSL instead\n");
+		fprintf(stderr, _("Set up DTLS failed; using SSL instead\n"));
 
-	vpninfo->progress(vpninfo, PRG_INFO,
-			  "Connected %s as %s%s%s, using %s\n", vpninfo->ifname,
-			  vpninfo->vpn_addr?:"",
-			  (vpninfo->vpn_addr6 && vpninfo->vpn_addr)?" + ":"",
-			  vpninfo->vpn_addr6?:"",
-			  (vpninfo->dtls_fd == -1) ?
-			      (vpninfo->deflate ? "SSL + deflate" : "SSL")
-			      : "DTLS");
+	vpn_progress(vpninfo, PRG_INFO,
+		     _("Connected %s as %s%s%s, using %s\n"), vpninfo->ifname,
+		     vpninfo->vpn_addr?:"",
+		     (vpninfo->vpn_addr6 && vpninfo->vpn_addr)?" + ":"",
+		     vpninfo->vpn_addr6?:"",
+		     (vpninfo->dtls_fd == -1) ?
+		     (vpninfo->deflate ? "SSL + deflate" : "SSL")
+		     : "DTLS");
 
 	if (!vpninfo->vpnc_script)
-		vpninfo->progress(vpninfo, PRG_INFO,
-				  "No --script argument provided; DNS and routing are not configured\n");
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("No --script argument provided; DNS and routing are not configured\n"));
 
 	if (background) {
 		int pid;
+
+		/* Open the pidfile before forking, so we can report errors
+		   more sanely. It's *possible* that we'll fail to write to
+		   it, but very unlikely. */
+		if (pidfile != NULL) {
+			fp = fopen(pidfile, "w");
+			if (!fp) {
+				fprintf(stderr, _("Failed to open '%s' for write: %s\n"),
+					pidfile, strerror(errno));
+				exit(1);
+			}
+		}
 		if ((pid = fork())) {
-			vpninfo->progress(vpninfo, PRG_INFO,
-					  "Continuing in background; pid %d\n",
-					  pid);
+			if (fp) {
+				fprintf(fp, "%d\n", pid);
+				fclose(fp);
+			}
+			vpn_progress(vpninfo, PRG_INFO,
+				     _("Continuing in background; pid %d\n"),
+				     pid);
 			exit(0);
 		}
+		if (fp)
+			fclose(fp);
 	}
 	vpn_mainloop(vpninfo);
+	if (fp)
+		unlink(pidfile);
 	exit(1);
 }
 
-static int write_new_config(struct openconnect_info *vpninfo, char *buf, int buflen)
+static int write_new_config(void *_vpninfo, char *buf, int buflen)
 {
+	struct openconnect_info *vpninfo = _vpninfo;
 	int config_fd;
 	int err;
 
 	config_fd = open(vpninfo->xmlconfig, O_WRONLY|O_TRUNC|O_CREAT, 0644);
 	if (config_fd < 0) {
 		err = errno;
-		fprintf(stderr, "Failed to open %s for write: %s\n",
+		fprintf(stderr, _("Failed to open %s for write: %s\n"),
 			vpninfo->xmlconfig, strerror(err));
 		return -err;
 	}
@@ -579,7 +653,7 @@ static int write_new_config(struct openconnect_info *vpninfo, char *buf, int buf
 	/* FIXME: We should actually write to a new tempfile, then rename */
 	if(write(config_fd, buf, buflen) != buflen) {
 		err = errno;
-		fprintf(stderr, "Failed to write config to %s: %s\n",
+		fprintf(stderr, _("Failed to write config to %s: %s\n"),
 			vpninfo->xmlconfig, strerror(err));
 
 		return -err;
@@ -588,7 +662,7 @@ static int write_new_config(struct openconnect_info *vpninfo, char *buf, int buf
 	return 0;
 }
 
-void write_progress(struct openconnect_info *info, int level, const char *fmt, ...)
+void write_progress(void *_vpninfo, int level, const char *fmt, ...)
 {
 	FILE *outf = level ? stdout : stderr;
 	va_list args;
@@ -600,8 +674,30 @@ void write_progress(struct openconnect_info *info, int level, const char *fmt, .
 	}
 }
 
-void syslog_progress(struct openconnect_info *info, int level,
-		     const char *fmt, ...)
+#ifdef ANDROID
+void syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
+{
+        static int l[4] = {
+		ANDROID_LOG_ERROR,	/* PRG_ERR   */
+		ANDROID_LOG_INFO,	/* PRG_INFO  */
+		ANDROID_LOG_DEBUG,	/* PRG_DEBUG */
+		ANDROID_LOG_DEBUG	/* PRG_TRACE */
+        };
+	va_list args, args2;
+
+	if (verbose >= level) {
+		va_start(args, fmt);
+		va_copy(args2, args);
+		__android_log_vprint(l[level], "openconnect", fmt, args);
+		/* Android wants it to stderr too, so the GUI can scrape
+		   it and display it as well as going to syslog */
+		vfprintf(stderr, fmt, args2);
+		va_end(args);
+		va_end(args2);
+	}
+}
+#else /* !ANDROID */
+void syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
 {
 	int priority = level ? LOG_INFO : LOG_NOTICE;
 	va_list args;
@@ -612,7 +708,7 @@ void syslog_progress(struct openconnect_info *info, int level,
 		va_end(args);
 	}
 }
-
+#endif
 
 struct accepted_cert {
 	struct accepted_cert *next;
@@ -620,9 +716,10 @@ struct accepted_cert {
 	char host[0];
 } *accepted_certs;
 
-static int validate_peer_cert(struct openconnect_info *vpninfo, X509 *peer_cert,
+static int validate_peer_cert(void *_vpninfo, X509 *peer_cert,
 			      const char *reason)
 {
+	struct openconnect_info *vpninfo = _vpninfo;
 	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
 	struct accepted_cert *this;
 	int ret;
@@ -645,12 +742,16 @@ static int validate_peer_cert(struct openconnect_info *vpninfo, X509 *peer_cert,
 		char buf[6];
 		int ret;
 
-		fprintf(stderr, "\nCertificate from VPN server \"%s\" failed verification.\n"
-			"Reason: %s\n",	vpninfo->hostname, reason);
+		fprintf(stderr,
+			_("\nCertificate from VPN server \"%s\" failed verification.\n"
+			  "Reason: %s\n"), vpninfo->hostname, reason);
+		if (non_inter)
+			return -EINVAL;
+
 		fflush(stderr);
 
 		ui = UI_new();
-		UI_add_input_string(ui, "Enter 'yes' to accept, 'no' to abort; anything else to view: ",
+		UI_add_input_string(ui, _("Enter 'yes' to accept, 'no' to abort; anything else to view: "),
 				    UI_INPUT_FLAG_ECHO, buf, 2, 5);
 		ret = UI_process(ui);
 		UI_free(ui);
@@ -659,7 +760,7 @@ static int validate_peer_cert(struct openconnect_info *vpninfo, X509 *peer_cert,
 		if (ret == -1)
 			buf[0] = 0;
 
-		if (!strcasecmp(buf, "yes")) {
+		if (!strcasecmp(buf, _("yes"))) {
 			struct accepted_cert *newcert = malloc(sizeof(*newcert) +
 							       strlen(vpninfo->hostname) + 1);
 			if (newcert) {
@@ -670,10 +771,11 @@ static int validate_peer_cert(struct openconnect_info *vpninfo, X509 *peer_cert,
 			}
 			return 0;
 		}
-		if (!strcasecmp(buf, "no"))
+		if (!strcasecmp(buf, _("no")))
 			return -EINVAL;
 
 		X509_print_fp(stderr, peer_cert);
+		fprintf(stderr, _("SHA1 fingerprint: %s\n"), fingerprint);
 	}
 				
 }
