@@ -27,7 +27,6 @@
 #include <limits.h>
 #include <sys/select.h>
 #include <signal.h>
-#include <arpa/inet.h>
 #include <unistd.h>
 #include <string.h>
 
@@ -80,13 +79,13 @@ int vpn_mainloop(struct openconnect_info *vpninfo)
 		struct timeval tv;
 		fd_set rfds, wfds, efds;
 
-#ifdef SSL_F_DTLS1_CONNECT
+#ifdef SSL_OP_CISCO_ANYCONNECT
 		if (vpninfo->new_dtls_ssl)
 			dtls_try_handshake(vpninfo);
 
 		if (vpninfo->dtls_attempt_period && !vpninfo->dtls_ssl && !vpninfo->new_dtls_ssl &&
 		    vpninfo->new_dtls_started + vpninfo->dtls_attempt_period < time(NULL)) {
-			vpninfo->progress(vpninfo, PRG_TRACE, "Attempt new DTLS connection\n");
+			vpn_progress(vpninfo, PRG_TRACE, _("Attempt new DTLS connection\n"));
 			connect_dtls_socket(vpninfo);
 		}
 		if (vpninfo->dtls_ssl)
@@ -118,8 +117,8 @@ int vpn_mainloop(struct openconnect_info *vpninfo)
 		if (did_work)
 			continue;
 
-		vpninfo->progress(vpninfo, PRG_TRACE,
-				  "No work to do; sleeping for %d ms...\n", timeout);
+		vpn_progress(vpninfo, PRG_TRACE,
+			     _("No work to do; sleeping for %d ms...\n"), timeout);
 		memcpy(&rfds, &vpninfo->select_rfds, sizeof(rfds));
 		memcpy(&wfds, &vpninfo->select_wfds, sizeof(wfds));
 		memcpy(&efds, &vpninfo->select_efds, sizeof(efds));
