@@ -22,7 +22,6 @@
  *   Boston, MA 02110-1301 USA
  */
 
-#define _GNU_SOURCE
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netdb.h>
@@ -40,6 +39,8 @@
 #include <sys/mount.h>
 #elif defined (__sun__) || defined(__NetBSD__) || defined(__DragonFly__)
 #include <sys/statvfs.h>
+#elif defined (__GNU__)
+#include <sys/statfs.h>
 #endif
 
 #include <openssl/ssl.h>
@@ -78,7 +79,7 @@ static int print_err(const char *str, size_t len, void *ptr)
 {
 	struct openconnect_info *vpninfo = ptr;
 
-	vpninfo->progress(vpninfo, PRG_ERR, "%s", str);
+	vpn_progress(vpninfo, PRG_ERR, "%s", str);
 	return 0;
 }
 
@@ -127,9 +128,9 @@ static int pem_pw_cb(char *buf, int len, int w, void *v)
 	SSL_CTX_set_default_passwd_cb_userdata(vpninfo->https_ctx, NULL);
 
 	if (len <= strlen(vpninfo->cert_password)) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "PEM password too long (%zd >= %d)\n",
-				  strlen(vpninfo->cert_password), len);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("PEM password too long (%zd >= %d)\n"),
+			     strlen(vpninfo->cert_password), len);
 		return -1;
 	}
 	strcpy(buf, vpninfo->cert_password);
@@ -163,12 +164,14 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 		if (ERR_GET_LIB(err) == ERR_LIB_PKCS12 &&
 		    ERR_GET_FUNC(err) == PKCS12_F_PKCS12_PARSE &&
 		    ERR_GET_REASON(err) == PKCS12_R_MAC_VERIFY_FAILURE) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Parse PKCS#12 failed (wrong passphrase?)\n");
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Parse PKCS#12 failed (wrong passphrase?)\n"));
 			vpninfo->cert_password = NULL;
 			goto retrypass;
 		}
 
-		vpninfo->progress(vpninfo, PRG_ERR, "Parse PKCS#12 failed (see above errors)\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Parse PKCS#12 failed (see above errors)\n"));
 		PKCS12_free(p12);
 		return -EINVAL;
 	}
@@ -176,8 +179,8 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 		vpninfo->cert_x509 = cert;
 		SSL_CTX_use_certificate(vpninfo->https_ctx, cert);
 	} else {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "PKCS#12 contained no certificate!");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("PKCS#12 contained no certificate!"));
 		ret = -EINVAL;
 	}
 
@@ -185,8 +188,8 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 		SSL_CTX_use_PrivateKey(vpninfo->https_ctx, pkey);
 		EVP_PKEY_free(pkey);
 	} else {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "PKCS#12 contained no private key!");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("PKCS#12 contained no private key!"));
 		ret = -EINVAL;
 	}
 
@@ -204,8 +207,8 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 
 				X509_NAME_oneline(X509_get_subject_name(cert2),
 						  buf, sizeof(buf));
-				vpninfo->progress(vpninfo, PRG_DEBUG,
-						  "Extra cert from PKCS#12: '%s'\n", buf);
+				vpn_progress(vpninfo, PRG_DEBUG,
+					     _("Extra cert from PKCS#12: '%s'\n"), buf);
 				CRYPTO_add(&cert2->references, 1, CRYPTO_LOCK_X509);
 				SSL_CTX_add_extra_chain_cert(vpninfo->https_ctx, cert2);
 				cert = cert2;
@@ -219,6 +222,7 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 	return ret;
 }
 
+#ifdef HAVE_ENGINE
 static int load_tpm_certificate(struct openconnect_info *vpninfo)
 {
 	ENGINE *e;
@@ -227,13 +231,13 @@ static int load_tpm_certificate(struct openconnect_info *vpninfo)
 
 	e = ENGINE_by_id("tpm");
 	if (!e) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Can't load TPM engine.\n");
+		vpn_progress(vpninfo, PRG_ERR, _("Can't load TPM engine.\n"));
 		report_ssl_errors(vpninfo);
 		return -EINVAL;
 	}
 	if (!ENGINE_init(e) || !ENGINE_set_default_RSA(e) ||
 	    !ENGINE_set_default_RAND(e)) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Failed to init TPM engine\n");
+		vpn_progress(vpninfo, PRG_ERR, _("Failed to init TPM engine\n"));
 		report_ssl_errors(vpninfo);
 		ENGINE_free(e);
 		return -EINVAL;
@@ -242,21 +246,22 @@ static int load_tpm_certificate(struct openconnect_info *vpninfo)
 	if (vpninfo->cert_password) {
 		if (!ENGINE_ctrl_cmd(e, "PIN", strlen(vpninfo->cert_password),
 				     vpninfo->cert_password, NULL, 0)) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Failed to set TPM SRK password\n");
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to set TPM SRK password\n"));
 			report_ssl_errors(vpninfo);
 		}
 	}
 	key = ENGINE_load_private_key(e, vpninfo->sslkey, NULL, NULL);
 	if (!key) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Failed to load TPM private key\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to load TPM private key\n"));
 		report_ssl_errors(vpninfo);
 		ENGINE_free(e);
 		ENGINE_finish(e);
 		return -EINVAL;
 	}
 	if (!SSL_CTX_use_PrivateKey(vpninfo->https_ctx, key)) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Add key from TPM failed\n");
+		vpn_progress(vpninfo, PRG_ERR, _("Add key from TPM failed\n"));
 		report_ssl_errors(vpninfo);
 		ENGINE_free(e);
 		ENGINE_finish(e);
@@ -264,6 +269,14 @@ static int load_tpm_certificate(struct openconnect_info *vpninfo)
 	}
 	return 0;
 }
+#else
+static int load_tpm_certificate(struct openconnect_info *vpninfo)
+{
+	vpn_progress(vpninfo, PRG_ERR,
+		     _("This version of OpenConnect was built without TPM support\n"));
+	return -EINVAL;
+}
+#endif
 
 static int reload_pem_cert(struct openconnect_info *vpninfo)
 {
@@ -275,8 +288,8 @@ static int reload_pem_cert(struct openconnect_info *vpninfo)
 	if (BIO_read_filename(b, vpninfo->cert) <= 0) {
 	err:
 		BIO_free(b);
-		vpninfo->progress(vpninfo, PRG_ERR,
-				 "Failed to reload X509 cert for expiry check\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to reload X509 cert for expiry check\n"));
 		report_ssl_errors(vpninfo);
 		return -EIO;
 	}
@@ -289,8 +302,8 @@ static int reload_pem_cert(struct openconnect_info *vpninfo)
 
 static int load_certificate(struct openconnect_info *vpninfo)
 {
-	vpninfo->progress(vpninfo, PRG_TRACE,
-			  "Using certificate file %s\n", vpninfo->cert);
+	vpn_progress(vpninfo, PRG_TRACE,
+		     _("Using certificate file %s\n"), vpninfo->cert);
 
 	if (vpninfo->cert_type == CERT_TYPE_PKCS12 ||
 	    vpninfo->cert_type == CERT_TYPE_UNKNOWN) {
@@ -299,9 +312,9 @@ static int load_certificate(struct openconnect_info *vpninfo)
 
 		f = fopen(vpninfo->cert, "r");
 		if (!f) {
-			vpninfo->progress(vpninfo, PRG_ERR,
-					  "Failed to open certificate file %s: %s\n",
-					  vpninfo->cert, strerror(errno));
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to open certificate file %s: %s\n"),
+				     vpninfo->cert, strerror(errno));
 			return -ENOENT;
 		}
 		p12 = d2i_PKCS12_fp(f, NULL);
@@ -311,7 +324,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 
 		/* Not PKCS#12 */
 		if (vpninfo->cert_type == CERT_TYPE_PKCS12) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Read PKCS#12 failed\n");
+			vpn_progress(vpninfo, PRG_ERR, _("Read PKCS#12 failed\n"));
 			report_ssl_errors(vpninfo);
 			return -EINVAL;
 		}
@@ -322,8 +335,8 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	/* It's PEM or TPM now, and either way we need to load the plain cert: */
 	if (!SSL_CTX_use_certificate_chain_file(vpninfo->https_ctx,
 						vpninfo->cert)) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Loading certificate failed\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Loading certificate failed\n"));
 		report_ssl_errors(vpninfo);
 		return -EINVAL;
 	}
@@ -336,9 +349,9 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		char buf[256];
 
 		if (!f) {
-			vpninfo->progress(vpninfo, PRG_ERR,
-					  "Failed to open private key file %s: %s\n",
-					  vpninfo->cert, strerror(errno));
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to open private key file %s: %s\n"),
+				     vpninfo->cert, strerror(errno));
 			return -ENOENT;
 		}
 
@@ -356,9 +369,9 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		}
 		fclose(f);
 		if (vpninfo->cert_type == CERT_TYPE_UNKNOWN) {
-			vpninfo->progress(vpninfo, PRG_ERR,
-					  "Failed to identify private key type in '%s'\n",
-					  vpninfo->sslkey);
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to identify private key type in '%s'\n"),
+				     vpninfo->sslkey);
 			return -EINVAL;
 		}
 	}
@@ -387,11 +400,13 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		if (ERR_GET_LIB(err) == ERR_LIB_EVP &&
 		    ERR_GET_FUNC(err) == EVP_F_EVP_DECRYPTFINAL_EX &&
 		    ERR_GET_REASON(err) == EVP_R_BAD_DECRYPT) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Loading private key failed (wrong passphrase?)\n");
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Loading private key failed (wrong passphrase?)\n"));
 			goto again;
 		}
 		
-		vpninfo->progress(vpninfo, PRG_ERR, "Loading private key failed (see above errors)\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Loading private key failed (see above errors)\n"));
 		return -EINVAL;
 	}
 	return 0;
@@ -435,8 +450,8 @@ static int check_server_cert(struct openconnect_info *vpninfo, X509 *cert)
 		return ret;
 
 	if (strcasecmp(vpninfo->servercert, fingerprint)) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Server SSL certificate didn't match: %s\n", fingerprint);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Server SSL certificate didn't match: %s\n"), fingerprint);
 		return -EINVAL;
 	}
 	return 0;
@@ -463,7 +478,14 @@ static int match_hostname_elem(const char *hostname, int helem_len,
 		return -1;
 	}
 
-	if (toupper(hostname[0]) == toupper(match[0]))
+	/* From the NetBSD (5.1) man page for ctype(3):
+           Values of type char or signed char must first be cast to unsigned char,
+	   to ensure that the values are within the correct range.  The result
+	   should then be cast to int to avoid warnings from some compilers.
+	   We do indeed get warning "array subscript has type 'char'" without
+	   the casts. Ick. */
+	if (toupper((int)(unsigned char)hostname[0]) ==
+	    toupper((int)(unsigned char)match[0]))
 		return match_hostname_elem(hostname + 1, helem_len - 1,
 					   match + 1, melem_len - 1);
 
@@ -503,7 +525,7 @@ static int match_hostname(const char *hostname, const char *match)
 }
 
 /* cf. RFC2818 and RFC2459 */
-int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
+static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
 {
 	STACK_OF(GENERAL_NAME) *altnames;
 	X509_NAME *subjname;
@@ -548,16 +570,16 @@ int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
 				continue;
 
 			if (!match_hostname(vpninfo->hostname, str)) {
-				vpninfo->progress(vpninfo, PRG_TRACE,
-						  "Matched DNS altname '%s'\n",
-						  str);
+				vpn_progress(vpninfo, PRG_TRACE,
+					     _("Matched DNS altname '%s'\n"),
+					     str);
 				GENERAL_NAMES_free(altnames);
 				OPENSSL_free(str);
 				return 0;
 			} else {
-				vpninfo->progress(vpninfo, PRG_TRACE,
-						  "No match for altname '%s'\n",
-						  str);
+				vpn_progress(vpninfo, PRG_TRACE,
+					     _("No match for altname '%s'\n"),
+					     str);
 			}
 			OPENSSL_free(str);
 		} else if (this->type == GEN_IPADD && addrlen) {
@@ -569,9 +591,9 @@ int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
 			} else if (this->d.ip->length == 16) {
 				family = AF_INET6;
 			} else {
-				vpninfo->progress(vpninfo, PRG_ERR,
-						  "Certificate has GEN_IPADD altname with bogus length %d\n",
-						  this->d.ip->length);
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Certificate has GEN_IPADD altname with bogus length %d\n"),
+					     this->d.ip->length);
 				continue;
 			}
 			
@@ -580,17 +602,17 @@ int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
 
 			if (this->d.ip->length == addrlen &&
 			    !memcmp(addrbuf, this->d.ip->data, addrlen)) {
-				vpninfo->progress(vpninfo, PRG_TRACE,
-						  "Matched IP%s address '%s'\n",
-						  (family == AF_INET6)?"v6":"",
-						  host);
+				vpn_progress(vpninfo, PRG_TRACE,
+					     _("Matched %s address '%s'\n"),
+					     (family == AF_INET6)?"IPv6":"IPv4",
+					     host);
 				GENERAL_NAMES_free(altnames);
 				return 0;
 			} else {
-				vpninfo->progress(vpninfo, PRG_TRACE,
-						  "No match for IP%s address '%s'\n",
-						  (family == AF_INET6)?"v6":"",
-						  host);
+				vpn_progress(vpninfo, PRG_TRACE,
+					     _("No match for %s address '%s'\n"),
+					     (family == AF_INET6)?"IPv6":"IPv4",
+					     host);
 			}
 		} else if (this->type == GEN_URI) {
 			char *str;
@@ -629,13 +651,14 @@ int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
 				goto no_uri_match;
 
 			if (url_path) {
-				vpninfo->progress(vpninfo, PRG_TRACE, "URI '%s' has non-empty path; ignoring\n",
-						  str);
+				vpn_progress(vpninfo, PRG_TRACE,
+					     _("URI '%s' has non-empty path; ignoring\n"),
+					     str);
 				goto no_uri_match_silent;
 			}
-			vpninfo->progress(vpninfo, PRG_TRACE,
-					  "Matched URI '%s'\n",
-					  str);
+			vpn_progress(vpninfo, PRG_TRACE,
+				     _("Matched URI '%s'\n"),
+				     str);
 			free(url_proto);
 			free(url_host);
 			free(url_path);
@@ -644,9 +667,9 @@ int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
 			return 0;
 
 		no_uri_match:
-			vpninfo->progress(vpninfo, PRG_TRACE,
-					  "No match for URI '%s'\n",
-					  str);
+			vpn_progress(vpninfo, PRG_TRACE,
+				     _("No match for URI '%s'\n"),
+				     str);
 		no_uri_match_silent:
 			free(url_proto);
 			free(url_host);
@@ -659,14 +682,16 @@ int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
 	/* According to RFC2818, we don't use the legacy subject name if
 	   there was an altname with DNS type. */
 	if (altdns) {
-		vpninfo->progress(vpninfo, PRG_ERR, "No altname in peer cert matched '%s'\n",
-				  vpninfo->hostname);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("No altname in peer cert matched '%s'\n"),
+			     vpninfo->hostname);
 		return -EINVAL;
 	}
 
 	subjname = X509_get_subject_name(peer_cert);
 	if (!subjname) {
-		vpninfo->progress(vpninfo, PRG_ERR, "No subject name in peer cert!\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("No subject name in peer cert!\n"));
 		return -EINVAL;
 	}
 
@@ -685,20 +710,21 @@ int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
 	i = ASN1_STRING_to_UTF8((void *)&subjstr, subjasn1);
 
 	if (!subjstr || strlen(subjstr) != i) {
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Failed to parse subject name in peer cert\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to parse subject name in peer cert\n"));
 		return -EINVAL;
 	}
 	ret = 0;
 
 	if (match_hostname(vpninfo->hostname, subjstr)) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Peer cert subject mismatch ('%s' != '%s')\n",
-				  subjstr, vpninfo->hostname);
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Peer cert subject mismatch ('%s' != '%s')\n"),
+			     subjstr, vpninfo->hostname);
 		ret = -EINVAL;
 	} else {
-		vpninfo->progress(vpninfo, PRG_TRACE,
-				  "Matched peer certificate subject name '%s'\n",
-				  subjstr);
+		vpn_progress(vpninfo, PRG_TRACE,
+			     _("Matched peer certificate subject name '%s'\n"),
+			     subjstr);
 	}
 
 	OPENSSL_free(subjstr);			  
@@ -723,15 +749,16 @@ static int verify_peer(struct openconnect_info *vpninfo, SSL *https_ssl)
 		if (vfy != X509_V_OK)
 			err_string = X509_verify_cert_error_string(vfy);
 		else if (match_cert_hostname(vpninfo, peer_cert))
-			err_string = "certificate does not match hostname";
+			err_string = _("certificate does not match hostname");
 
 		if (err_string) {
-			vpninfo->progress(vpninfo, PRG_ERR,
-					  "Server certificate verify failed: %s\n",
-					  err_string);
+			vpn_progress(vpninfo, PRG_INFO,
+				     _("Server certificate verify failed: %s\n"),
+				     err_string);
 
 			if (vpninfo->validate_peer_cert)
-				ret = vpninfo->validate_peer_cert(vpninfo, peer_cert,
+				ret = vpninfo->validate_peer_cert(vpninfo->cbdata,
+								  peer_cert,
 								  err_string);
 			else
 				ret = -EINVAL;
@@ -773,8 +800,8 @@ static void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
 		cert = cert2;
 		X509_NAME_oneline(X509_get_subject_name(cert),
 				  buf, sizeof(buf));
-		vpninfo->progress(vpninfo, PRG_DEBUG,
-				  "Extra cert from cafile: '%s'\n", buf);
+		vpn_progress(vpninfo, PRG_DEBUG,
+			     _("Extra cert from cafile: '%s'\n"), buf);
 		SSL_CTX_add_extra_chain_cert(vpninfo->https_ctx, cert);
 	}
 	X509_STORE_CTX_cleanup(&ctx);
@@ -793,9 +820,10 @@ static int ssl_app_verify_callback(X509_STORE_CTX *ctx, void *arg)
 static int check_certificate_expiry(struct openconnect_info *vpninfo)
 {
 	ASN1_TIME *notAfter;
-	char *reason = NULL;
+	const char *reason = NULL;
 	time_t t;
 	int i;
+
 	if (!vpninfo->cert_x509)
 		return 0;
 
@@ -803,21 +831,22 @@ static int check_certificate_expiry(struct openconnect_info *vpninfo)
 	notAfter = X509_get_notAfter(vpninfo->cert_x509);
 	i = X509_cmp_time(notAfter, &t);
 	if (!i) {
-		vpninfo->progress(vpninfo, PRG_ERR, "Error in client cert notAfter field\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Error in client cert notAfter field\n"));
 		return -EINVAL;
 	} else if (i < 0) {
-		reason = "has expired";
+		reason = _("Client certificate has expired at");
 	} else {
-		t += 60 * 86400;
+		t += vpninfo->cert_expire_warning;
 		i = X509_cmp_time(notAfter, &t);
 		if (i < 0) {
-			reason = "expires soon";
+			reason = _("Client certificate expires soon at");
 		}
 	}
 	if (reason) {
 		BIO *bp = BIO_new(BIO_s_mem());
 		BUF_MEM *bm;
-		char *expiry = "<error>";
+		const char *expiry = _("<error>");
 		char zero = 0;
 
 		if (bp) {
@@ -826,8 +855,7 @@ static int check_certificate_expiry(struct openconnect_info *vpninfo)
 			BIO_get_mem_ptr(bp, &bm);
 			expiry = bm->data;
 		}
-		vpninfo->progress(vpninfo, PRG_ERR, "Client certificate %s at: %s\n",
-				  reason, expiry);
+		vpn_progress(vpninfo, PRG_ERR, "%s: %s\n", reason, expiry);
 		if (bp)
 			BIO_free(bp);
 	}
@@ -849,9 +877,15 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		ssl_sock = socket(vpninfo->peer_addr->sa_family, SOCK_STREAM, IPPROTO_IP);
 		if (ssl_sock < 0) {
 		reconn_err:
-			vpninfo->progress(vpninfo, PRG_ERR, "Failed to reconnect to %s %s\n",
-					  vpninfo->proxy?"proxy":"host",
-					  vpninfo->proxy?:vpninfo->hostname);
+			if (vpninfo->proxy) {
+				vpn_progress(vpninfo, PRG_ERR, 
+					     _("Failed to reconnect to proxy %s\n"),
+					     vpninfo->proxy);
+			} else {
+				vpn_progress(vpninfo, PRG_ERR, 
+					     _("Failed to reconnect to host %s\n"),
+					     vpninfo->hostname);
+			}
 			return -EINVAL;
 		}
 		if (connect(ssl_sock, vpninfo->peer_addr, vpninfo->peer_addrlen))
@@ -875,7 +909,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		   this way than if we pass NULL to getaddrinfo() and
 		   then try to fill in the numeric value into
 		   different types of returned sockaddr_in{6,}. */
-#ifdef OPENCONNECT_LIBPROXY
+#ifdef LIBPROXY_HDR
 		if (vpninfo->proxy_factory) {
 			char *url;
 			char **proxies;
@@ -911,8 +945,9 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			free(url);
 			free(proxies);
 			if (vpninfo->proxy)
-				vpninfo->progress(vpninfo, PRG_TRACE, "Proxy from libproxy: %s://%s:%d/\n",
-						  vpninfo->proxy_type, vpninfo->proxy, vpninfo->port);
+				vpn_progress(vpninfo, PRG_TRACE,
+					     _("Proxy from libproxy: %s://%s:%d/\n"),
+					     vpninfo->proxy_type, vpninfo->proxy, vpninfo->port);
 		}
 #endif
 		if (vpninfo->proxy) {
@@ -941,8 +976,9 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			free(hostname);
 
 		if (err) {
-			vpninfo->progress(vpninfo, PRG_ERR, "getaddrinfo failed for host '%s': %s\n",
-					  hostname, gai_strerror(err));
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("getaddrinfo failed for host '%s': %s\n"),
+				     hostname, gai_strerror(err));
 			return -EINVAL;
 		}
 
@@ -951,12 +987,12 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 
 			if (!getnameinfo(rp->ai_addr, rp->ai_addrlen, host,
 					 sizeof(host), NULL, 0, NI_NUMERICHOST))
-				vpninfo->progress(vpninfo, PRG_INFO,
-						  "Attempting to connect to %s%s%s:%s\n",
-						  rp->ai_family == AF_INET6?"[":"",
-						  host,
-						  rp->ai_family == AF_INET6?"]":"",
-						  port);
+				vpn_progress(vpninfo, PRG_INFO,
+					     _("Attempting to connect to %s%s%s:%s\n"),
+					     rp->ai_family == AF_INET6?"[":"",
+					     host,
+					     rp->ai_family == AF_INET6?"]":"",
+					     port);
 			
 			ssl_sock = socket(rp->ai_family, rp->ai_socktype,
 					  rp->ai_protocol);
@@ -967,7 +1003,8 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 				   use it again later */
 				vpninfo->peer_addr = malloc(rp->ai_addrlen);
 				if (!vpninfo->peer_addr) {
-					vpninfo->progress(vpninfo, PRG_ERR, "Failed to allocate sockaddr storage\n");
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Failed to allocate sockaddr storage\n"));
 					close(ssl_sock);
 					return -ENOMEM;
 				}
@@ -981,8 +1018,9 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		freeaddrinfo(result);
 		
 		if (ssl_sock < 0) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Failed to connect to host %s\n",
-					  vpninfo->proxy?:vpninfo->hostname);
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to connect to host %s\n"),
+				     vpninfo->proxy?:vpninfo->hostname);
 			return -EINVAL;
 		}
 	}
@@ -996,15 +1034,21 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		}
 	}
 
-	ssl3_method = SSLv3_client_method();
+	ssl3_method = TLSv1_client_method();
 	if (!vpninfo->https_ctx) {
 		vpninfo->https_ctx = SSL_CTX_new(ssl3_method);
 
+		/* Some servers (or their firewalls) really don't like seeing
+		   extensions. */
+#ifdef SSL_OP_NO_TICKET
+		SSL_CTX_set_options (vpninfo->https_ctx, SSL_OP_NO_TICKET);
+#endif
+
 		if (vpninfo->cert) {
 			err = load_certificate(vpninfo);
 			if (err) {
-				vpninfo->progress(vpninfo, PRG_ERR,
-						  "Loading certificate failed. Aborting.\n");
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Loading certificate failed. Aborting.\n"));
 				return err;
 			}
 			check_certificate_expiry(vpninfo);
@@ -1026,8 +1070,9 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 
 		if (vpninfo->cafile) {
 			if (!SSL_CTX_load_verify_locations(vpninfo->https_ctx, vpninfo->cafile, NULL)) {
-				vpninfo->progress(vpninfo, PRG_ERR, "Failed to open CA file '%s'\n",
-						  vpninfo->cafile);
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Failed to open CA file '%s'\n"),
+					     vpninfo->cafile);
 				report_ssl_errors(vpninfo);
 				close(ssl_sock);
 				return -EINVAL;
@@ -1041,11 +1086,11 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	https_bio = BIO_new_socket(ssl_sock, BIO_NOCLOSE);
 	SSL_set_bio(https_ssl, https_bio, https_bio);
 
-	vpninfo->progress(vpninfo, PRG_INFO,
-			  "SSL negotiation with %s\n", vpninfo->hostname);
+	vpn_progress(vpninfo, PRG_INFO, _("SSL negotiation with %s\n"),
+		     vpninfo->hostname);
 
 	if (SSL_connect(https_ssl) <= 0) {
-		vpninfo->progress(vpninfo, PRG_ERR, "SSL connection failure\n");
+		vpn_progress(vpninfo, PRG_ERR, _("SSL connection failure\n"));
 		report_ssl_errors(vpninfo);
 		SSL_free(https_ssl);
 		close(ssl_sock);
@@ -1061,8 +1106,8 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	vpninfo->ssl_fd = ssl_sock;
 	vpninfo->https_ssl = https_ssl;
 
-	vpninfo->progress(vpninfo, PRG_INFO,
-			  "Connected to HTTPS on %s\n", vpninfo->hostname);
+	vpn_progress(vpninfo, PRG_INFO, _("Connected to HTTPS on %s\n"),
+		     vpninfo->hostname);
 
 	return 0;
 }
@@ -1093,7 +1138,8 @@ int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 
 	if (statvfs(vpninfo->sslkey, &buf)) {
 		int err = errno;
-		vpninfo->progress(vpninfo, PRG_ERR, "statvfs: %s\n", strerror(errno));
+		vpn_progress(vpninfo, PRG_ERR, _("statvfs: %s\n"),
+			     strerror(errno));
 		return -err;
 	}
 	if (asprintf(&vpninfo->cert_password, "%lx", buf.f_fsid))
@@ -1109,7 +1155,8 @@ int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 
 	if (statfs(vpninfo->sslkey, &buf)) {
 		int err = errno;
-		vpninfo->progress(vpninfo, PRG_ERR, "statfs: %s\n", strerror(errno));
+		vpn_progress(vpninfo, PRG_ERR, _("statfs: %s\n"),
+			     strerror(errno));
 		return -err;
 	}
 	fsid64 = ((unsigned long long)fsid[0] << 32) | fsid[1];
