@@ -79,8 +79,8 @@
 static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
 {
 	if (!mtu_only)
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "No vpnc-script configured. Need Solaris IP-setting code\n");
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("No vpnc-script configured. Need Solaris IP-setting code\n"));
 	return 0;
 }
 #else
@@ -91,7 +91,7 @@ static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
 
 	net_fd = socket(PF_INET, SOCK_DGRAM, 0);
 	if (net_fd < 0) {
-		perror("open net");
+		perror(_("open net"));
 		return -EINVAL;
 	}
 	memset(&ifr, 0, sizeof(ifr));
@@ -101,22 +101,22 @@ static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
 		struct sockaddr_in addr;
 
 		if (ioctl(net_fd, SIOCGIFFLAGS, &ifr) < 0)
-			perror("SIOCGIFFLAGS");
+			perror(_("SIOCGIFFLAGS"));
 
 		ifr.ifr_flags |= IFF_UP | IFF_POINTOPOINT;
 		if (ioctl(net_fd, SIOCSIFFLAGS, &ifr) < 0)
-			perror("SIOCSIFFLAGS");
+			perror(_("SIOCSIFFLAGS"));
 
 		addr.sin_family = AF_INET;
 		addr.sin_addr.s_addr = inet_addr(vpninfo->vpn_addr);
 		memcpy(&ifr.ifr_addr, &addr, sizeof(addr));
 		if (ioctl(net_fd, SIOCSIFADDR, &ifr) < 0)
-			perror("SIOCSIFADDR");
+			perror(_("SIOCSIFADDR"));
 	}
 
 	ifr.ifr_mtu = vpninfo->mtu;
 	if (ioctl(net_fd, SIOCSIFMTU, &ifr) < 0)
-		perror("SIOCSIFMTU");
+		perror(_("SIOCSIFMTU"));
 
 	close(net_fd);
 
@@ -143,19 +143,25 @@ static int netmasklen(struct in_addr addr)
 }
 
 static int process_split_xxclude(struct openconnect_info *vpninfo,
-				 char *in_ex, char *route, int *v4_incs,
+				 int include, const char *route, int *v4_incs,
 				 int *v6_incs)
 {
 	struct in_addr addr;
+	const char *in_ex = include?"IN":"EX";
 	char envname[80];
 	char *slash;
 
 	slash = strchr(route, '/');
 	if (!slash) {
 	badinc:
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Discard bad split %sclude: \"%s\"\n",
-				  in_ex, route);
+		if (include)
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Discard bad split include: \"%s\"\n"),
+				     route);
+		else
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Discard bad split exclude: \"%s\"\n"),
+				     route);
 		return -EINVAL;
 	}
 
@@ -250,7 +256,8 @@ static void set_banner(struct openconnect_info *vpninfo)
 	q = banner;
 	
 	while (*p) {
-		if (*p == '%' && isxdigit(p[1]) && isxdigit(p[2])) {
+		if (*p == '%' && isxdigit((int)(unsigned char)p[1]) &&
+		    isxdigit((int)(unsigned char)p[2])) {
 			*(q++) = unhex(p + 1);
 			p += 3;
 		} else 
@@ -332,7 +339,7 @@ static void set_script_env(struct openconnect_info *vpninfo)
 		int nr_v6_split_includes = 0;
 
 		while (this) {
-			process_split_xxclude(vpninfo, "IN", this->route,
+			process_split_xxclude(vpninfo, 1, this->route,
 					      &nr_split_includes,
 					      &nr_v6_split_includes);
 			this = this->next;
@@ -348,7 +355,7 @@ static void set_script_env(struct openconnect_info *vpninfo)
 		int nr_v6_split_excludes = 0;
 
 		while (this) {
-			process_split_xxclude(vpninfo, "EX", this->route,
+			process_split_xxclude(vpninfo, 0, this->route,
 					      &nr_split_excludes,
 					      &nr_v6_split_excludes);
 			this = this->next;
@@ -365,14 +372,19 @@ static int script_config_tun(struct openconnect_info *vpninfo)
 {
 	if (system(vpninfo->vpnc_script)) {
 		int e = errno;
-		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Failed to spawn script '%s': %s\n",
-				  vpninfo->vpnc_script, strerror(e));
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to spawn script '%s': %s\n"),
+			     vpninfo->vpnc_script, strerror(e));
 		return -e;
 	}
 	return 0;
 }
 
+void script_reconnect (struct openconnect_info *vpninfo)
+{
+	setenv("reason", "reconnect", 1);
+	script_config_tun(vpninfo);
+}
 
 /* Set up a tuntap device. */
 int setup_tun(struct openconnect_info *vpninfo)
@@ -386,24 +398,24 @@ int setup_tun(struct openconnect_info *vpninfo)
 		int fds[2];
 
 		if (socketpair(AF_UNIX, SOCK_DGRAM, 0, fds)) {
-			perror("socketpair");
+			perror(_("socketpair"));
 			exit(1);
 		}
 		tun_fd = fds[0];
 		child = fork();
 		if (child < 0) {
-			perror("fork");
+			perror(_("fork"));
 			exit(1);
 		} else if (!child) {
 			close(tun_fd);
 			setenv_int("VPNFD", fds[1]);
 			execl("/bin/sh", "/bin/sh", "-c", vpninfo->vpnc_script, NULL);
-			perror("execl");
+			perror(_("execl"));
 			exit(1);
 		}
 		close(fds[1]);
 		vpninfo->script_tun = child;
-		vpninfo->ifname = "(script)";
+		vpninfo->ifname = strdup(_("(script)"));
 	} else {
 #ifdef IFF_TUN /* Linux */
 		struct ifreq ifr;
@@ -423,9 +435,9 @@ int setup_tun(struct openconnect_info *vpninfo)
 			if (errno != -ENOENT)
 				tunerr = errno;
 
-			vpninfo->progress(vpninfo, PRG_ERR,
-					  "Failed to open tun device: %s\n",
-					  strerror(tunerr));
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to open tun device: %s\n"),
+				     strerror(tunerr));
 			exit(1);
 		}
 		memset(&ifr, 0, sizeof(ifr));
@@ -434,9 +446,9 @@ int setup_tun(struct openconnect_info *vpninfo)
 			strncpy(ifr.ifr_name, vpninfo->ifname,
 				sizeof(ifr.ifr_name) - 1);
 		if (ioctl(tun_fd, TUNSETIFF, (void *) &ifr) < 0) {
-			vpninfo->progress(vpninfo, PRG_ERR,
-					  "TUNSETIFF failed: %s\n",
-					  strerror(errno));
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("TUNSETIFF failed: %s\n"),
+				     strerror(errno));
 			exit(1);
 		}
 		if (!vpninfo->ifname)
@@ -448,20 +460,20 @@ int setup_tun(struct openconnect_info *vpninfo)
 		struct ifreq ifr;
 
 		if (ip_fd < 0) {
-			perror("open /dev/ip");
+			perror(_("open /dev/ip"));
 			return -EIO;
 		}
 
 		tun_fd = open("/dev/tun", O_RDWR);
 		if (tun_fd < 0) {
-			perror("open /dev/tun");
+			perror(_("open /dev/tun"));
 			close(ip_fd);
 			return -EIO;
 		}
 
 		unit_nr = ioctl(tun_fd, TUNNEWPPA, -1);
 		if (unit_nr < 0) {
-			perror("Failed to create new tun");
+			perror(_("Failed to create new tun"));
 			close(tun_fd);
 			close(ip_fd);
 			return -EIO;
@@ -469,20 +481,20 @@ int setup_tun(struct openconnect_info *vpninfo)
 		
 		tun2_fd = open("/dev/tun", O_RDWR);
 		if (tun2_fd < 0) {
-			perror("open /dev/tun again");
+			perror(_("open /dev/tun again"));
 			close(tun_fd);
 			close(ip_fd);
 			return -EIO;
 		}
 		if (ioctl(tun2_fd, I_PUSH, "ip") < 0) {
-			perror("Can't push IP");
+			perror(_("Can't push IP"));
 			close(tun2_fd);
 			close(tun_fd);
 			close(ip_fd);
 			return -EIO;
 		}
 		if (ioctl(tun2_fd, IF_UNITSEL, &unit_nr) < 0) {
-			perror("Can't select unit");
+			perror(_("Can't select unit"));
 			close(tun2_fd);
 			close(tun_fd);
 			close(ip_fd);
@@ -490,7 +502,7 @@ int setup_tun(struct openconnect_info *vpninfo)
 		}
 		mux_id = ioctl(ip_fd, I_PLINK, tun2_fd);
 		if (mux_id < 0) {
-			perror("Can't link tun to IP");
+			perror(_("Can't link tun to IP"));
 			close(tun2_fd);
 			close(tun_fd);
 			close(ip_fd);
@@ -499,21 +511,21 @@ int setup_tun(struct openconnect_info *vpninfo)
 		close(tun2_fd);
 
 		sprintf(tun_name, "tun%d", unit_nr);
-		vpninfo->ifname = tun_name;
+		vpninfo->ifname = strdup(tun_name);
 
 		memset(&ifr, 0, sizeof(ifr));
 		strcpy(ifr.ifr_name, tun_name);
 		ifr.ifr_ip_muxid = mux_id;
 
 		if (ioctl(ip_fd, SIOCSIFMUXID, &ifr) < 0) {
-			perror("Set mux id");
+			perror(_("Set mux id"));
 			close(tun_fd);
 			ioctl(ip_fd, I_PUNLINK, mux_id);
 			close(ip_fd);
 			return -EIO;
 		}
 		/* Solaris tunctl needs this in order to tear it down */
-		vpninfo->progress(vpninfo, PRG_DEBUG, "mux id is %d\n", mux_id);
+		vpn_progress(vpninfo, PRG_DEBUG, _("mux id is %d\n"), mux_id);
 		vpninfo->tun_muxid = mux_id;
 		vpninfo->ip_fd = ip_fd;
 
@@ -527,14 +539,14 @@ int setup_tun(struct openconnect_info *vpninfo)
 				break;
 		}
 		if (tun_fd < 0) {
-			perror("open tun");
+			perror(_("open tun"));
 			exit(1);
 		}
-		vpninfo->ifname = tun_name + 5;
+		vpninfo->ifname = strdup(tun_name + 5);
 #ifdef TUNSIFHEAD
 		i = 1;
 		if (ioctl(tun_fd, TUNSIFHEAD, &i) < 0) {
-			perror("TUNSIFHEAD");
+			perror(_("TUNSIFHEAD"));
 			exit(1);
 		}
 #endif
@@ -567,10 +579,39 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	unsigned char buf[2000];
 	int len;
 	int work_done = 0;
+	int moredata = 1;
 
 	if (FD_ISSET(vpninfo->tun_fd, &vpninfo->select_rfds)) {
-		while ((len = read(vpninfo->tun_fd, buf, sizeof(buf))) > 0) {
+		while (moredata) {
 			unsigned char *pkt = buf;
+#ifdef __sun__
+			/* On Solaris, if we're actually using tuntap and *not* just
+			   passing packets through a UNIX socket to a child process,
+			   we have to use getmsg() to ensure that we only get *one*
+			   packet at a time. */
+			if (!vpninfo->script_tun) {
+				struct strbuf strb;
+				int flags = 0;
+				int ret;
+				
+				strb.buf = buf;
+				strb.maxlen = sizeof(buf);
+				strb.len = 0;
+				ret = getmsg(vpninfo->tun_fd, NULL, &strb, &flags);
+				if (ret < 0)
+					break;
+
+				if (!(ret & MOREDATA))
+					moredata = 0;
+				len = strb.len;
+			} else
+#endif /* __sun__ ... */
+			{
+				len = read(vpninfo->tun_fd, buf, sizeof(buf));
+				if (len <= 0)
+					break;
+			}
+
 #ifdef TUN_HAS_AF_PREFIX
 			if (!vpninfo->script_tun) {
 				pkt += 4;
@@ -613,9 +654,9 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 				static int complained = 0;
 				if (!complained) {
 					complained = 1;
-					vpninfo->progress(vpninfo, PRG_ERR,
-							  "Unknown packet (len %d) received: %02x %02x %02x %02x...\n",
-							  len, data[0], data[1], data[2], data[3]);
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Unknown packet (len %d) received: %02x %02x %02x %02x...\n"),
+						     len, data[0], data[1], data[2], data[3]);
 				}
 				free(this);
 				continue;
@@ -646,15 +687,15 @@ void shutdown_tun(struct openconnect_info *vpninfo)
 		if (vpninfo->vpnc_script) {
 			setenv("reason", "disconnect", 1);
 			if (system(vpninfo->vpnc_script) == -1) {
-				vpninfo->progress(vpninfo, PRG_ERR,
-						  "Failed to spawn script '%s': %s\n",
-						  vpninfo->vpnc_script,
-						  strerror(errno));
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Failed to spawn script '%s': %s\n"),
+					     vpninfo->vpnc_script,
+					     strerror(errno));
 			}
 		}
 #ifdef __sun__
 		if (ioctl(vpninfo->ip_fd, I_PUNLINK, vpninfo->tun_muxid) < 0)
-			perror("ioctl(I_PUNLINK)");
+			perror(_("ioctl(I_PUNLINK)"));
 
 		close(vpninfo->ip_fd);
 		vpninfo->ip_fd = -1;
