@@ -1,8 +1,8 @@
 #!/bin/sh
 
-v="v3.02"
+v="v3.15"
 
-if tag=`git describe --tags`; then
+if [ -d .git ] && tag=`git describe --tags`; then
 	v="$tag"
 
 	# Update the index from working tree first
@@ -16,5 +16,5 @@ else
 	v="$v"-unknown
 fi
 
-echo "char openconnect_version[] = \"$v\";" > version.c
+echo "char openconnect_version[] = \"$v\";" > $1
 echo "New version: $v"
